import os
import urllib.error
import urllib.request
from tempfile import NamedTemporaryFile

import pytest

import genomepy.online
from tests import linux, travis


def test_download_file():
    # HTTP
    tmp = NamedTemporaryFile().name
    assert not os.path.exists(tmp)
    url = "http://hgdownload.soe.ucsc.edu/goldenPath/ailMel1/bigZips/md5sum.txt"
    genomepy.online.download_file(url, tmp)
    assert os.path.exists(tmp)

    # FTP (doesn't work on Travis-Linux)
    if not (travis and linux):
        tmp = NamedTemporaryFile().name
        assert not os.path.exists(tmp)
        url = "ftp://ftp.ncbi.nlm.nih.gov//genomes/all/GCF/000/027/325/GCF_000027325.1_ASM2732v1/README.txt"
        genomepy.online.download_file(url, tmp)
        assert os.path.exists(tmp)


def test_connect_ftp_link():
    if not (travis and linux):  # (doesn't work on Travis-Linux)
        # good FTP host
        ftp_link = "ftp://ftp.ncbi.nlm.nih.gov/genomes/README.txt"
        ftp, target = genomepy.online.connect_ftp_link(ftp_link)
        assert target == "genomes/README.txt"
        result = ftp.nlst(target)
        ftp.quit()  # logout
        assert result == [target]

        # bad FTP host
        with pytest.raises(genomepy.exceptions.GenomeDownloadError):
            genomepy.online.connect_ftp_link("ftp://not.an.ftp/at/all")


def test_read_url(
    url="http://ftp.xenbase.org/pub/Genomics/JGI/README", expected="The data"
):
    text = genomepy.online.read_url(url)
    assert text.startswith(expected)


def test_retry(capsys):
    # runs passed function
    txt = "hello world"
    genomepy.online.retry(print, 1, txt)
    captured = capsys.readouterr().out.strip()
    assert captured == txt

    # handles URLErrors
    def _offline_func():
        raise urllib.error.URLError("this function is offline")

    assert genomepy.online.retry(_offline_func, 1) is None


def test_check_url():
    # good URL
    assert genomepy.online.check_url(
        "https://mbdata.science.ru.nl/siebrenf/GRCh37/GRCh37/README.txt"
    )
    assert genomepy.online.check_url("http://ftp.xenbase.org/pub/Genomics/JGI/README")

    # bad URL:
    assert not genomepy.online.check_url("https://www.thiswebsiteisoffline.nl/")

    if not (travis and linux):  # (doesn't work on Travis-Linux)
        # good FTP
        ftp_link = (
            "ftp://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/027/325/"
            "GCF_000027325.1_ASM2732v1/GCF_000027325.1_ASM2732v1_genomic.gff.gz"
        )
        assert genomepy.online.check_url(ftp_link)

        # bad FTP target
        assert not genomepy.online.check_url("ftp://ftp.ncbi.nlm.nih.gov/bad_target")



import os
from tempfile import NamedTemporaryFile

import pyfaidx
import pytest

import genomepy.files
import genomepy.utils
from genomepy.seq import as_seqdict


# to ignore file changes
# git update-index --assume-unchanged tests/data/small_genome.fa.gz
# to recheck file changes
# git update-index --no-assume-unchanged tests/data/small_genome.fa.gz
def test_genome__init__(small_genome):
    # genome dir not found
    with pytest.raises(FileNotFoundError):
        genomepy.Genome("unknown", "unknown")

    # no fasta file
    with pytest.raises(FileNotFoundError):
        genomepy.Genome("empty", "tests/data")

    readme = "tests/data/README.txt"
    if os.path.exists(readme):
        os.unlink(readme)

    assert small_genome.genomes_dir == genomepy.utils.get_genomes_dir(None, False)
    assert small_genome.name == "small_genome"
    assert small_genome.filename == os.path.abspath("tests/data/small_genome.fa.gz")
    assert small_genome.genome_dir == os.path.dirname(small_genome.filename)
    assert os.path.exists(small_genome.index_file)
    assert os.path.exists(small_genome.sizes_file)
    assert os.path.exists(small_genome.gaps_file)
    assert isinstance(small_genome.sizes, dict)
    assert isinstance(small_genome.gaps, dict)
    assert small_genome.annotation_gtf_file is None
    assert small_genome.annotation_bed_file is None
    assert small_genome.tax_id == small_genome.assembly_accession == "na"
    assert isinstance(small_genome.plugin, dict)


def test_sizes(gap_genome):
    assert list(gap_genome.sizes.keys()) == ["chr1", "chr2", "chr3"]
    assert all(
        isinstance(gap_genome.sizes[chrom], int) for chrom in gap_genome.sizes.keys()
    )
    assert gap_genome.sizes["chr1"] == 28

    # does not overwrite user-set sizes
    gap_genome.sizes = {"asd": 1}
    assert gap_genome.sizes == {"asd": 1}

    # repopulates empty dicts
    gap_genome.sizes = None
    assert list(gap_genome.sizes.keys()) == [
        "chr1",
        "chr2",
        "chr3",
    ]  # noqa: it's a lazy attribute


def test_gaps(gap_genome):
    assert list(gap_genome.gaps.keys()) == ["chr1", "chr3"]

    # does not overwrite user-set gaps
    gap_genome.gaps = {"asd": 1}
    assert gap_genome.gaps == {"asd": 1}

    # repopulates empty dicts
    gap_genome.gaps = None
    assert list(gap_genome.gaps.keys()) == [
        "chr1",
        "chr3",
    ]  # noqa: it's a lazy attribute


# def test__parse_name(small_genome):
#     # name
#     name = small_genome._parse_name("test")
#     assert name == "test"
#
#     # file
#     name = small_genome._parse_name("/home/genomepy/genomes/test2.fa")
#     assert name == "test2"
#
#     # url
#     name = small_genome._parse_name(
#         "http://ftp.xenbase.org/pub/Genomics/JGI/Xentr9.1/XT9_1.fa.gz"
#     )
#     assert name == "XT9_1"


def test__parse_filename(small_genome):
    gpath = "tests/data/small_genome.fa.gz"

    # file path
    filename = small_genome._parse_filename(gpath)
    assert filename == os.path.abspath(gpath)

    # folder path
    filename = small_genome._parse_filename(os.path.dirname(gpath))
    assert filename == os.path.abspath(gpath)

    # name of genome in genomes_dir
    os.mkdir("tests/data/small_genome")
    with open("tests/data/small_genome/small_genome.fa.gz", "w") as fa:
        fa.write("test")
    small_genome.genomes_dir = "tests/data/"
    filename = small_genome._parse_filename(os.path.basename(gpath))
    assert filename == "tests/data/small_genome/small_genome.fa.gz"
    genomepy.utils.rm_rf("tests/data/small_genome")

    # genome not found
    with pytest.raises(FileNotFoundError):
        small_genome._parse_filename("does not exist")


def test__check_annotation_file(small_genome):
    # does not exist
    gtf = small_genome._check_annotation_file("gtf")
    assert gtf is None

    # does exist
    path = "tests/data/small_genome.annotation.test.gz"
    with open(path, "w") as fa:
        fa.write("test")
    test = small_genome._check_annotation_file("test")
    assert test == os.path.abspath(path)
    os.unlink(path)


def test_generate_gap_bed(fname="tests/data/gap.fa", outname="tests/data/gap.bed"):
    tmp = NamedTemporaryFile().name
    genomepy.genome.generate_gap_bed(fname, tmp)

    result = open(tmp).read()
    expected = open(outname).read()

    assert result == expected
    genomepy.files.rm_rf(fname + ".fai")


def test_generate_fa_sizes(fname="tests/data/gap.fa"):
    tmp = NamedTemporaryFile().name
    genomepy.genome.generate_fa_sizes(fname, tmp)

    result = open(tmp).read()

    assert result == "chr1\t28\nchr2\t45\nchr3\t15\n"
    genomepy.files.rm_rf(fname + ".fai")


# genome.sequences.py


def test_get_track_type():
    tracks = [
        (("chr1:10-20", "chr2:10-20"), "interval"),
        (["chr1:10-20", "chr2:10-20"], "interval"),
        ("tests/data/regions.txt", "interval"),
        ("tests/data/regions2.txt", "interval"),
        ("tests/data/regions.bed", "bed"),
        ("tests/data/regions2.bed", "bed"),
    ]
    for track, track_type in tracks:
        result = genomepy.genome.sequences.get_track_type(track)
        assert result == track_type


def test_region_to_seq(small_genome, region="chrI:10-20"):
    # extract sequences marked in track from small_genome.fa.gz
    seq = genomepy.genome.sequences.region_to_seq(
        small_genome, region=region, extend_up=0, extend_down=0
    )
    assert seq == "CCCACACACC"

    # extend by varying amounts
    seq = genomepy.genome.sequences.region_to_seq(
        small_genome, region=region, extend_up=1, extend_down=2
    )
    # original:    "CCCACACACC"
    assert seq == "ACCCACACACCCA"


def test__regions_to_seqs(small_genome, track="tests/data/regions2.txt"):
    # extract sequences marked in regions.bed from small_genome.fa.gz
    seqs = genomepy.genome.sequences.regions_to_seqs(
        small_genome, track=track, extend_up=0, extend_down=0
    )
    for i, seq in enumerate(seqs):
        assert seq.name == ["chrI:10-20", "chrII:20-30"][i]
        assert seq.seq == ["CCCACACACC", "TCCTCCAAGC"][i]

    # extend by varying amounts
    seqs = genomepy.genome.sequences.regions_to_seqs(
        small_genome, track=track, extend_up=1, extend_down=2
    )
    for i, seq in enumerate(seqs):
        assert seq.name == ["chrI:10-20", "chrII:20-30"][i]
        # original:         "CCCACACACC",    "TCCTCCAAGC"
        assert seq.seq == ["ACCCACACACCCA", "CTCCTCCAAGCCC"][i]


def test_bed_to_seqs(small_genome, track="tests/data/regions.bed"):
    # extract sequences marked in regions.bed from small_genome.fa.gz
    seqs = genomepy.genome.sequences.bed_to_seqs(
        small_genome, track=track, stranded=False, extend_up=0, extend_down=0
    )
    for i, seq in enumerate(seqs):
        assert seq.name == ["chrI:10-20 gene_a", "chrII:20-30 gene_b"][i]
        assert seq.seq == ["CCCACACACC", "TCCTCCAAGC"][i]

    # BED3 (no gene names)
    seqs = genomepy.genome.sequences.bed_to_seqs(
        small_genome,
        track="tests/data/regions2.bed",
        stranded=False,
        extend_up=0,
        extend_down=0,
    )
    for i, seq in enumerate(seqs):
        assert seq.name == ["chrI:10-20", "chrII:20-30"][i]
        assert seq.seq == ["CCCACACACC", "TCCTCCAAGC"][i]

    # second sequence is on the negative strand
    seqs = genomepy.genome.sequences.bed_to_seqs(
        small_genome, track=track, stranded=True, extend_up=0, extend_down=0
    )
    for i, seq in enumerate(seqs):
        assert seq.name == ["chrI:10-20 gene_a", "chrII:20-30 gene_b"][i]
        # original:        "CCCACACACC", "TCCTCCAAGC"
        assert seq.seq == ["CCCACACACC", "GCTTGGAGGA"][i]

    # extend by varying amounts
    seqs = genomepy.genome.sequences.bed_to_seqs(
        small_genome, track=track, stranded=True, extend_up=1, extend_down=2
    )
    for i, seq in enumerate(seqs):
        assert seq.name == ["chrI:10-20 gene_a", "chrII:20-30 gene_b"][i]
        # original:         "CCCACACACC",    "GCTTGGAGGA"
        assert seq.seq == ["ACCCACACACCCA", "GGCTTGGAGGAGA"][i]


def test_track2fasta(small_genome):
    tracks = [
        ("tests/data/regions.txt", "interval"),
        ("tests/data/regions.bed", "bed"),
    ]
    for i, track in enumerate(tracks):
        seq = small_genome.track2fasta(
            track=track[0],
            fastafile=None,
            stranded=False,
            extend_up=i,
            extend_down=i + 1,
        )
        # default sequence:       CCCACACACC
        if i == 0:  # extend up +0, down -1
            assert seq[0].seq == "CCCACACACCC"
            assert seq[1].seq == "TCCTCCAAGCC"
        else:  # extend up +1, down -4
            assert seq[0].seq == "ACCCACACACCCA"
            assert seq[1].seq == "CTCCTCCAAGCCC"


def test_weighted_selection(n=2):
    tuples = [(1, "one"), (2, "two"), (3, "three")]
    ws = genomepy.genome.sequences.weighted_selection(tuples, n)

    assert len(ws) == n
    assert isinstance(ws[0], str)
    assert tuples[0][1] in ws or tuples[1][1] in ws or tuples[2][1] in ws


def test_get_random_sequences(small_genome):
    n = 2
    length = 200  # default
    chroms = ["chrI", "chrII"]
    max_n = 0.1  # default
    rs = small_genome.get_random_sequences(
        n=n, length=length, chroms=chroms, max_n=max_n
    )

    # check that the output has the right length, content, types, and sequence length
    assert len(rs) == n
    for i in range(n):
        assert rs[i][0] in chroms
        assert (
            isinstance(rs[i][0], str)
            and isinstance(rs[i][1], int)
            and isinstance(rs[i][2], int)
        )
        assert rs[i][2] - rs[i][1] == length

    # check that the max Ns are lower than the expected cutoff
    rs = small_genome.get_random_sequences(n=1, chroms=chroms, outtype="string")
    assert (
        str(small_genome.track2fasta(rs[0])[0].seq).upper().count("N") <= length * max_n
    )


# seq.py


def test_as_seqdict(small_genome):
    test_data = [
        "tests/data/as_seqdict/test.bed",
        "tests/data/as_seqdict/test.fa",
        "tests/data/as_seqdict/test.fasta",
        "tests/data/as_seqdict/test.txt",
        # pybedtools.BedTool("tests/data/as_seqdict/test.bed"),
        ["chrI:110-120", "chrII:130-140", "chrIII:410-420"],
        # np.array(['chrI:110-120', 'chrII:130-140', 'chrIII:410-420']),
        pyfaidx.Fasta("tests/data/as_seqdict/test.fa"),
    ]

    # test different inputs
    for dataset in test_data:
        result = as_seqdict(dataset, genome=small_genome.filename)
        assert "chrI:110-120" in result, "key not present"
        assert "chrII:130-140" in result, "key not present"
        assert "chrIII:410-420" in result, "key not present"
        assert result["chrI:110-120"] == "CTCTCAACTT", "sequence incorrect"
        assert result["chrII:130-140"] == "TGTCTCTCGC", "sequence incorrect"
        assert result["chrIII:410-420"] == "TCCCAACTTA", "sequence incorrect"

    # test minsize argument
    for dataset in test_data:
        with pytest.raises(ValueError):
            as_seqdict(dataset, genome="tests/data/small_genome.fa.gz", minsize=100)

    # raise error on empty file
    with pytest.raises(IOError):
        as_seqdict("tests/data/as_seqdict/empty.fa")

    # test genome@chrom:start-end format
    datasets = [
        "tests/data/as_seqdict/test.with_genome.txt",
        [
            "tests/data/small_genome.fa.gz@chrI:110-120",
            "tests/data/small_genome.fa.gz@chrII:130-140",
            "tests/data/small_genome.fa.gz@chrIII:410-420",
        ],
    ]

    for dataset in datasets:
        assert sorted(as_seqdict(dataset).values()) == [
            "CTCTCAACTT",
            "TCCCAACTTA",
            "TGTCTCTCGC",
        ]



import pytest

import genomepy
from tests import linux, travis


@pytest.mark.skipif(travis and linux, reason="FTP does not work on Travis-Linux")
def test_gencodeprovider(gencode):
    assert gencode.name == "GENCODE"
    assert gencode.taxid_fields == ["taxonomy_id"]


@pytest.mark.skipif(travis and linux, reason="FTP does not work on Travis-Linux")
def test_genome_info_tuple(gencode):
    t = gencode._genome_info_tuple("GRCh37")
    assert isinstance(t, tuple)
    assert t[0:4] == ("GRCh37", "GCA_000001405.1", 9606, True)


@pytest.mark.skipif(travis and linux, reason="FTP does not work on Travis-Linux")
def test_genomes(gencode):
    assert gencode.genomes["GRCh37"]["other_info"] == "GENCODE annotation + UCSC genome"
    assert gencode.genomes["GRCh38"]["assembly_accession"] == "GCA_000001405.15"


@pytest.mark.skipif(travis and linux, reason="FTP does not work on Travis-Linux")
def test_get_genome_download_link(gencode):
    link = gencode.get_genome_download_link("GRCh37", mask="soft")
    assert link in [
        "http://hgdownload.soe.ucsc.edu/goldenPath/hg19/bigZips/chromFa.tar.gz",
        "http://hgdownload.soe.ucsc.edu/goldenPath/hg19/bigZips/hg19.fa.masked.gz",
    ]


@pytest.mark.skipif(travis and linux, reason="FTP does not work on Travis-Linux")
def test_get_annotation_download_links(gencode):
    # default annotation filing system
    genome = "GRCm39"
    annots = gencode.get_annotation_download_links(genome)
    expected = [  # release numbers removed
        "ftp.ebi.ac.uk/pub/databases/gencode/Gencode_mouse/release_M",
        "/gencode.vM",
        ".annotation.gtf.gz",
    ]
    assert all([exp in annots[0] for exp in expected])

    # GRCh37, the one with the unique filing system.
    genome = "GRCh37"
    annots = gencode.get_annotation_download_links(genome)
    expected = [  # release numbers removed
        "ftp.ebi.ac.uk/pub/databases/gencode/Gencode_human/release_",
        "/GRCh37_mapping/gencode.v",
        "lift37.annotation.gtf.gz",
    ]
    assert all([exp in annots[0] for exp in expected])


@pytest.mark.skipif(travis and linux, reason="FTP does not work on Travis-Linux")
def test_download_annotation(gencode):
    gencode.download_annotation("GRCm39")  # smallest gencode annotation (0.8 GB)


def test_get_gencode2ucsc():
    genomes = {
        "test1": {"species": "Homo sapiens"},
        "test2": {"species": "Mus Musculus"},
        "test3": {"species": "whatever"},
    }
    gencode2ucsc = genomepy.providers.gencode.get_gencode2ucsc(genomes)
    assert gencode2ucsc["test1"] == "hg1"
    assert gencode2ucsc["test2"] == "mm2"
    assert gencode2ucsc["test3"] == "mm3"


def test_get_releases():
    listing = [
        "/path/to/release_22",
        "/path/to/mouse/release_M44",
        "/path/to/mouse/release_M33",
        "/path/to/release_01",  # too old
        "/path/to/something/else",
    ]
    specie = "human"
    releases = genomepy.providers.gencode.get_releases(listing, specie)

    assert releases == ["44", "33", "22"]

    specie = "mouse"
    releases = genomepy.providers.gencode.get_releases(listing, specie)

    assert releases == ["M44", "M33", "M22"]


def test_add_grch37():
    release = 42
    genomes = {
        "GRCh11": {},
        "GRCh22": {"annotations": [f"ftp/to/release_{release}/gtf"]},
    }
    genomes = genomepy.providers.gencode.add_grch37(genomes, "")
    expected = (
        f"/Gencode_human/release_{release}/GRCh37_mapping/"
        f"gencode.v{release}lift37.annotation.gtf.gz"
    )

    assert genomes["GRCh22"]["annotations"] == [f"ftp/to/release_{release}/gtf"]
    assert genomes["GRCh37"]["annotations"] == [expected]



import os

import pandas as pd
import pytest

import genomepy
import genomepy.annotation.sanitize
import genomepy.files
import genomepy.utils
from genomepy.annotation import query_mygene


def test__get_name_and_dir():
    # working inputs
    t_dicts = [
        {
            "name": "GRCz11",
            "genomes_dir": "tests/data",
            "expected_name": "GRCz11",
            "expected_dir": "tests/data/GRCz11",
        },
        {
            "name": "tests/data/GRCz11",
            "genomes_dir": None,
            "expected_name": "GRCz11",
            "expected_dir": "tests/data/GRCz11",
        },
        {
            "name": "tests/data/GRCz11/GRCz11.annotation.bed",
            "genomes_dir": None,
            "expected_name": "GRCz11",
            "expected_dir": "tests/data/GRCz11",
        },
        # works with warnings
        {
            "name": "empty",
            "genomes_dir": "tests/data",
            "expected_name": "empty",
            "expected_dir": "tests/data/empty",
        },
    ]
    for td in t_dicts:
        name, genome_dir = genomepy.annotation._get_name_and_dir(
            td["name"], td["genomes_dir"]
        )
        assert name == td["expected_name"], td
        assert genome_dir == genomepy.utils.cleanpath(td["expected_dir"]), td

    # does not exist
    with pytest.raises(FileNotFoundError):
        genomepy.annotation._get_name_and_dir("tests/data/GRCz11/what.who")
    # no bed/gtf/fa
    with pytest.raises(NotImplementedError):
        genomepy.annotation._get_name_and_dir("tests/data/GRCz11/README.txt")


def test_annotation_init(caplog, annot):
    assert annot.genome_file.endswith("data/regexp/regexp.fa")
    assert annot.annotation_gtf_file.endswith("data/regexp/regexp.annotation.gtf")
    assert annot.annotation_bed_file.endswith("data/regexp/regexp.annotation.bed")

    assert annot.annotation_contigs == ["chrM"]  # from BED
    assert len(annot.genome_contigs) == 17  # from sizes
    assert isinstance(annot.bed, pd.DataFrame)
    assert isinstance(annot.gtf, pd.DataFrame)
    assert annot.genes("bed") == ["NP_059343.1"]

    # >1 GTF files: take unzipped
    g1 = "tests/data/data.annotation.gtf"
    with genomepy.files._open(g1, "w") as fa:
        fa.write("1")
    g2 = "tests/data/data.annotation.gtf.gz"
    with genomepy.files._open(g2, "w") as fa:
        fa.write("2")
    a = genomepy.Annotation("data", genomes_dir="tests")
    assert a.annotation_gtf_file.endswith(g1)
    genomepy.utils.rm_rf(g1)
    genomepy.utils.rm_rf(g2)

    # not enough GTF files
    genomepy.Annotation("empty", genomes_dir="tests/data")
    assert "Could not find 'empty.annotation.bed" in caplog.text
    assert "Could not find 'empty.annotation.gtf" in caplog.text

    # Genome doesn't exist
    with pytest.raises(FileNotFoundError):
        genomepy.Annotation("never_existed", genomes_dir="tests/data")


def test_custom_annotation():
    for fname in [
        "tests/data/custom.annotation.bed",
        "tests/data/custom.annotation.bed.gz",
    ]:
        a = genomepy.Annotation(name=fname, genomes_dir="tests/data")
        assert a.bed.shape[0] == 10

    for fname in [
        "tests/data/custom.annotation.gtf",
        "tests/data/custom.annotation.gtf.gz",
    ]:
        a = genomepy.Annotation(name=fname, genomes_dir="tests/data")
        assert a.gtf.shape[0] == 45

    with pytest.raises(NotImplementedError):
        a = genomepy.Annotation(name="tests/data/regions.txt", genomes_dir="tests/data")


def test_named_gtf():
    a = genomepy.Annotation("sacCer3", genomes_dir="tests/data")
    df = a.named_gtf
    assert df.index.name == "gene_name"
    assert str(a.gtf.index.dtype) == "int64"
    assert str(df.index.dtype) == "object"
    assert set(df.at["YDL248W", "seqname"]) == {"chrIV"}


def test_genes():
    a = genomepy.Annotation("GRCz11", genomes_dir="tests/data")
    g = a.genes("bed")
    assert isinstance(g, list)
    g = a.genes("gtf")
    assert isinstance(g, list)


def test_gene_coords(caplog):
    a = genomepy.Annotation("sacCer3", genomes_dir="tests/data")

    bed_genes = a.genes("bed")[0:10]
    c = a.gene_coords(bed_genes, "bed")
    assert list(c.shape) == [10, 5]
    assert c.columns.to_list() == ["chrom", "start", "end", "name", "strand"]

    gtf_genes = a.genes("gtf")[0:10]
    c = a.gene_coords(gtf_genes, "gtf")
    assert list(c.shape) == [10, 5]
    assert c.columns.to_list() == ["seqname", "start", "end", "gene_name", "strand"]

    # mismatched gene names!
    _ = a.gene_coords(["what?"], "bed")
    assert "No genes found." in caplog.text

    _ = a.gene_coords(["YDL200C_mRNA", "what?"], "bed")
    assert "Only 50% of genes was found." in caplog.text


def test_map_locations():
    a = genomepy.Annotation("sacCer3", genomes_dir="tests/data")

    # BED + Ensembl
    ens = a.map_locations(annot=a.bed.head(1), to="Ensembl")
    assert ens.chrom.to_list() == ["IV"]

    # GTF + NCBI
    ncbi = a.map_locations(annot=a.gtf.head(1), to="NCBI")
    assert ncbi.seqname.to_list() == ["IV"]

    # custom dataframe (already indexed)
    df = a.bed.head(1).set_index("start")
    cus = a.map_locations(annot=df, to="Ensembl")
    assert ens.chrom.to_list() == ["IV"]
    assert cus.index.name == "start"


def test_filter_regex():
    # match
    gtf_file = "tests/data/regexp/regexp.annotation.gtf"
    with open(gtf_file, "w") as f:
        f.write("chr2\tcool_gene\n")
        f.write("chromosome3\tboring_gene\n")
    a = genomepy.annotation.Annotation("regexp", genomes_dir="tests/data")
    gtf_df = pd.read_csv(gtf_file, sep="\t", header=None)

    df = a.filter_regex(gtf_df, regex=".*2")
    assert isinstance(df, pd.DataFrame)
    assert df[0].to_list() == ["chr2"]

    # invert match
    df = a.filter_regex(gtf_df, regex=".*2", invert_match=True)
    assert isinstance(df, pd.DataFrame)
    assert df[0].to_list() == ["chromosome3"]

    genomepy.utils.rm_rf(gtf_file)


def test_gtf_dict():
    a = genomepy.annotation.Annotation("GRCz11", genomes_dir="tests/data")
    gid2gname = a.gtf_dict("gene_id", "gene_name")
    assert gid2gname["ENSDARG00000103202"] == "CR383668.1"
    gid2gname = a.gtf_dict("gene_id", "gene_name", False)
    assert gid2gname["ENSDARG00000103202"] == ["CR383668.1"]

    seq2tid = a.gtf_dict("seqname", "transcript_id")
    assert seq2tid["1"] == ["ENSDART00000171868"]
    #
    # with pytest.raises(IndexError):
    #     a.gtf_dict("gene_name", "transcript_id", annot="bed")


def test_lengths():
    a = genomepy.annotation.Annotation("GRCz11", genomes_dir="tests/data")

    lengths = a.lengths("gene_name")
    assert str(lengths.dtype) == "uint32"
    expected_genes = a.named_gtf[a.named_gtf.feature == "exon"].index.unique()
    assert sorted(expected_genes) == sorted(lengths.index)
    assert lengths.loc["CR383668.1"] == 420

    lengths = a.lengths("gene_id")
    assert lengths.index[0].startswith("ENSDARG")
    assert lengths.loc["ENSDARG00000103202"] == 420

    lengths = a.lengths("transcript_name")
    assert lengths.loc["CR383668.1-201"] == 420

    lengths = a.lengths("transcript_id")
    assert lengths.index[0].startswith("ENSDART")
    assert lengths.loc["ENSDART00000159919"] == 420


# annotation.utils.py


def test_count_columns():
    cols = genomepy.annotation.utils.count_columns(
        "tests/data/sacCer3/sacCer3.annotation.gtf"
    )
    assert cols == 9

    cols = genomepy.annotation.utils.count_columns(
        "tests/data/sacCer3/sacCer3.annotation.bed"
    )
    assert cols == 12


def test_read_annot():
    df = genomepy.annotation.utils.read_annot(
        "tests/data/sacCer3/sacCer3.annotation.bed"
    )
    assert len(df.columns) == 12
    assert list(df.columns).__eq__(list(genomepy.annotation.utils.BED12_FORMAT))


def test_write_annot():
    for ext in ["bed", "gtf"]:
        df1 = genomepy.annotation.utils.read_annot(
            f"tests/data/sacCer3/sacCer3.annotation.{ext}"
        )
        genomepy.annotation.utils.write_annot(
            df1, f"tests/data/sacCer3/test.annotation.{ext}"
        )
        df2 = genomepy.annotation.utils.read_annot(
            f"tests/data/sacCer3/test.annotation.{ext}"
        )

        os.remove(f"tests/data/sacCer3/test.annotation.{ext}")
        assert df1.equals(df2)


def test_generate_annot():
    bed_file = "tests/data/data.annotation.bed.gz"
    with genomepy.files._open(bed_file, "w") as f:
        f.write("this is the old bed\n")

    gtf_file = "tests/data/data.annotation.gtf.gz"
    with genomepy.files._open(gtf_file, "w") as f:
        f.write(
            "chr1\tgenomepy\texon\t1\t100\t.\t+\t.\t"
            'gene_id "ENSGP1234"; transcript_id "GP_1234.1";  gene_name "GP";\n'
        )
    a = genomepy.annotation.Annotation("data", genomes_dir="tests")

    # BED file before & after
    with genomepy.files._open(a.annotation_bed_file, "r") as f:
        lines = f.readlines()
    assert lines == ["this is the old bed\n"]

    genomepy.annotation.utils.generate_annot(
        a.annotation_gtf_file, a.annotation_bed_file, True
    )
    with genomepy.files._open(a.annotation_bed_file, "r") as f:
        lines = f.readlines()
    assert lines == ["chr1\t0\t100\tGP_1234.1\t0\t+\t100\t100\t0\t1\t100,\t0,\n"]
    assert len(lines) == 1

    # unzipping and rezipping correctly
    assert a.annotation_gtf_file.endswith(".gz")
    assert a.annotation_bed_file.endswith(".gz")

    # does not overwrite by default
    with pytest.raises(FileExistsError):
        genomepy.annotation.utils.generate_annot(
            a.annotation_gtf_file, a.annotation_bed_file
        )

    genomepy.utils.rm_rf(gtf_file)
    genomepy.utils.rm_rf(bed_file)


def test__parse_annot():
    a = genomepy.Annotation("sacCer3", genomes_dir="tests/data")
    df = genomepy.annotation.utils._parse_annot(a, "bed")
    assert df.equals(a.bed)
    df = genomepy.annotation.utils._parse_annot(a, "gtf")
    assert df.equals(a.gtf)
    df = genomepy.annotation.utils._parse_annot(a, a.bed)
    assert df.equals(a.bed)

    with pytest.raises(ValueError):
        genomepy.annotation.utils._parse_annot(a, 1)


# annotation.sanitize.py


def test_match_contigs():
    # nothing to work with
    a = genomepy.Annotation("sacCer3", genomes_dir="tests/data")
    cd = genomepy.annotation.sanitize._match_contigs(a)
    assert cd is None

    # one missing contig, one fixable contig
    a = genomepy.Annotation("sanitize", genomes_dir="tests/data")
    before = a.gtf
    cd = genomepy.annotation.sanitize._match_contigs(a)
    after = a.gtf

    assert cd == {"NC_007112.7": "1"}
    assert a.genome_contigs == ["1"]
    assert list(before.seqname.unique()) == ["NC_007112.7", "NC_002333.2"]
    assert list(after.seqname.unique()) == ["1", "NC_002333.2"]


def test_filter_contigs():
    a = genomepy.Annotation("sacCer3", genomes_dir="tests/data")
    assert "chrV" not in a.bed.chrom.unique()

    # add a chromosome to the BED not present in the genome
    a.bed.at[0, "chrom"] = "chrV"
    assert "chrV" in a.bed.chrom.unique()

    missing_contigs = genomepy.annotation.sanitize._filter_contigs(a)
    assert missing_contigs == {"chrV"}
    assert "chrV" not in a.bed.chrom.unique()


def test_document_sanitizing():
    class HasReadme:
        readme_file = "tests/data/test.log"

        def __init__(self):
            with open(self.readme_file, "w"):
                pass

        def _check_property(self):
            pass

    a = HasReadme()
    cd = {"1": "a", "2": "b"}
    mc = ["1", "2", "3", "4", "5"]
    genomepy.annotation.sanitize._document_sanitizing(a, cd, mc)
    md, lines = genomepy.files.read_readme(a.readme_file)
    genomepy.utils.rm_rf(a.readme_file)

    status = md["sanitized annotation"]
    assert status == "2 contigs were renamed. 5 contigs were removed (see below)."
    assert lines == [
        "",
        "The following contigs were filtered out of the gene annotation:",
        "1, 2, 3, 4, 5",
    ]


def test_sanitize():
    a = genomepy.Annotation("sanitize", genomes_dir="tests/data")

    a.sanitize(overwrite=False)
    assert a.gtf.shape == (4, 9)
    assert list(set(a.gtf.seqname)) == ["1"]


# annotation.mygene.py


def test_map_genes():
    a = genomepy.Annotation("GRCz11", genomes_dir="tests/data")

    bed = a.bed.head()
    transcript_ids = bed.name.to_list()
    assert transcript_ids[0] == "ENSDART00000159919"

    # transcript to gene
    res = a.map_genes(field="ensembl.gene", annot=bed)
    genes = res.name.to_list()
    assert genes[0] == "ENSDARG00000103202"

    # # transcript to symbol
    # res = a.map_genes(field="symbol", df=bed)
    # symbol = res.name.to_list()
    # assert symbol[0] == "CR383668.1"

    # refseq hits & subtypes
    protein = a.map_genes(field="refseq", product="protein", annot=bed)
    assert protein.name.to_list()[0].startswith("NP_")
    # rna = a.map_genes(field="refseq", product="rna", df=bed)
    # assert rna.name.to_list()[0].startswith("NM_")
    # assert rna.shape == protein.shape


def test_query_mygene():
    result = query_mygene(["ENST00000449992"], 9606, "symbol")
    assert result.loc["ENST00000449992", "symbol"] == "TP63"


def test_parse_mygene_input():
    # wrong product
    with pytest.raises(ValueError):
        genomepy.annotation.mygene._parse_mygene_input("", "illegal!")

    # wrong field
    with pytest.raises(ValueError):
        genomepy.annotation.mygene._parse_mygene_input("illegal!", None)

    # correct input
    gf, p = genomepy.annotation.mygene._parse_mygene_input(
        "refseq.translation.rna", None
    )
    assert gf == "refseq.translation.rna" and p is None
    gf, p = genomepy.annotation.mygene._parse_mygene_input("NAME", "RNA")
    assert gf == "name" and p == "rna"


# def test_ensembl_genome_info():
#     # Works
#     a = genomepy.Annotation("sacCer3", genomes_dir="tests/data")
#     egi = genomepy.annotation.mygene.ensembl_genome_info(a)
#     assert egi == ("R64-1-1", "GCA_000146045.2", 4932)
#     # genomepy.utils.rm_rf(os.path.join(a.genome_dir, "assembly_report.txt"))
#
#     # No readme
#     a = genomepy.Annotation("regexp", genomes_dir="tests/data")
#     readme_file = os.path.join(a.genome_dir, "README.txt")
#     with pytest.raises(FileNotFoundError):
#         genomepy.annotation.mygene.ensembl_genome_info(a)
#
#     # Empty readme
#     with open(readme_file, "w") as f:
#         f.write("\n")
#     a = genomepy.Annotation("regexp", genomes_dir="tests/data")
#     egi = genomepy.annotation.mygene.ensembl_genome_info(a)
#     assert egi is None
#     genomepy.utils.rm_rf(readme_file)



"""
Genes and genomes at your fingertips!
"""

import sys

from loguru import logger

from genomepy.__about__ import __author__, __version__
from genomepy.annotation import Annotation, query_mygene
from genomepy.caching import clean
from genomepy.config import manage_config
from genomepy.functions import (
    head_annotations,
    install_genome,
    list_available_genomes,
    list_installed_genomes,
)
from genomepy.genome import Genome
from genomepy.plugins import manage_plugins
from genomepy.providers import Provider, list_online_providers, list_providers, search

# Public API objects
# everything made available with `from genomepy import *`
__all__ = [
    # modules
    "genome",
    "annotation",
    "providers",
    "files",
    "online",
    "utils",
    "exceptions",
    # objects
    "__author__",
    "__version__",
    "Genome",
    "Annotation",
    "Provider",
    "search",
    "head_annotations",
    "install_genome",
    "list_providers",
    "list_online_providers",
    "list_installed_genomes",
    "list_available_genomes",
    "manage_plugins",
    "manage_config",
    "clean",
    "query_mygene",
]

# No traceback
# sys.tracebacklimit = 0

# logger is a singleton, configuration here will be used module-wide
logger.remove()
logger.add(
    sys.stderr,
    format="<green>{time:HH:mm:ss}</green> <bold>|</bold> <blue>{level}</blue> <bold>|</bold> {message}",
    level="INFO",
)



import os
from tempfile import mkdtemp
from time import sleep

import pytest

import genomepy
import genomepy.utils
from tests import linux
from tests.conftest import validate_annot

skip = False
if not skip:

    @pytest.fixture(scope="module", params=["no-overwrite", "overwrite"])
    def force(request):
        return request.param

    @pytest.fixture(scope="module", params=["original_name", "use_localname"])
    def localname(request):
        return request.param

    def test_install_genome_options(
        force, localname, genome="ASM2732v1", provider="NCBI"
    ):
        """Test force, localname and bgzip"""
        tmp = mkdtemp()
        force = False if force == "no-overwrite" else True
        localname = None if localname == "original_name" else "My_localname"

        genomepy.install_genome(
            genome,
            provider,
            genomes_dir=tmp,
            localname=localname,
            force=False,
        )
        sleep(1)

        # force test
        name = genomepy.utils.get_localname(genome, localname)
        path = os.path.join(tmp, name, name + ".fa")

        t0 = os.path.getmtime(path)
        # OSX rounds down getmtime to the second
        if not linux:
            sleep(1)
        genomepy.install_genome(
            genome,
            provider,
            genomes_dir=tmp,
            localname=localname,
            force=force,
        )
        sleep(1)

        t1 = os.path.getmtime(path)
        assert t0 != t1 if force else t0 == t1

        genomepy.utils.rm_rf(tmp)

    def test_install_annotation_options(
        force, localname, genome="ASM14646v1", provider="NCBI"
    ):
        """Test force and localname with annotations"""
        tmp = mkdtemp()
        force = False if force == "no-overwrite" else True
        localname = None if localname == "original_name" else "My_localname"

        # create dummy fasta to skip download_genome step
        name = genomepy.utils.get_localname(genome, localname)
        path = os.path.join(tmp, name, name + ".fa")
        os.mkdir(os.path.dirname(path))
        with open(path, "w") as f:
            f.write(">Chr1\nAAAACCCCTTTTGGGG\n")
        genomepy.install_genome(
            genome,
            provider,
            genomes_dir=tmp,
            localname=localname,
            only_annotation=True,
            skip_matching=True,
            skip_filter=True,
            force=False,
        )
        sleep(1)

        gtf = os.path.join(tmp, name, name + ".annotation.gtf")
        validate_annot(gtf, "gtf")

        bed = os.path.join(tmp, name, name + ".annotation.bed")
        validate_annot(bed, "bed")

        # force test
        t0 = os.path.getmtime(gtf)
        # OSX rounds down getmtime to the second
        if not linux:
            sleep(1)
        genomepy.install_genome(
            genome,
            provider,
            genomes_dir=tmp,
            localname=localname,
            only_annotation=True,
            skip_matching=True,
            skip_filter=True,
            force=force,
        )
        sleep(1)

        t1 = os.path.getmtime(gtf)
        assert t0 != t1 if force else t0 == t1

        genomepy.utils.rm_rf(tmp)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

project = 'genomepy'
copyright = 'Siebren Frölich, Maarten van der Sande, Tilman Schäfers and Simon van Heeringen'
author = 'Siebren Frölich, Maarten van der Sande, Tilman Schäfers and Simon van Heeringen'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',      # automatic documentation from docstrings
    'sphinx.ext.coverage',     # gather documentation coverage stats
    'sphinx.ext.napoleon',     # recognize numpy & google style docstrings
    'sphinx.ext.autosummary',  # Create neat summary tables
    'm2r2',                    # recognize markdown files
]

# Configuration of sphinx.ext.autosummary
autosummary_generate = True

# Configuration of m2r2
source_suffix = ['.rst', '.md']

# Configuration of sphinx.ext.coverage
coverage_show_missing_items = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# add the global ToC to the sidebar of each page
html_sidebars = {
    "**": ["globaltoc.html", "relations.html", "sourcelink.html", "searchbox.html"]
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_last_updated_fmt = '%Y-%m-%d, %H:%M (UTC)'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# add the "Edit on GitHub" link to the docs
html_context = {
    "display_github": True,
    "github_host": "github.com",
    "github_user": "vanheeringen-lab",
    "github_repo": "genomepy",
    "github_version": "develop",
    "conf_py_path": "/docs/",
    "source_suffix": [".rst", ".md"],
}



import os
import subprocess as sp
from shutil import rmtree

import norns
import pandas as pd
import pytest
from appdirs import user_cache_dir
from diskcache import Cache

import genomepy

from . import linux, travis


def test_linting():
    out = sp.check_output(
        "chmod +x tests/format.sh; tests/format.sh lint",
        stderr=sp.STDOUT,  # send errors to out
        shell=True,
    )
    out = out.decode("utf-8").replace("\x1b[0m", "").replace("\nDone\n", "")
    if out != "":
        pytest.fail(
            f"Linting failed. Messages: \n\n{out}\n"
            "Run `tests/format.sh` to format the repo.",
            False,
        )


# caching.py


@pytest.mark.skipif(not travis, reason="it works locally all right")
def test_clean():
    genomepy.clean()
    my_cache_dir = os.path.join(user_cache_dir("genomepy"), genomepy.__version__)
    assert os.path.exists(my_cache_dir)  # dir exists
    assert not os.listdir(my_cache_dir)  # contains 0 pickles
    genomepy.clean()  # no errors when cache dir is empty


@pytest.mark.skipif(not travis, reason="it works locally all right")
def test_cache():
    # test caching of complex data types
    my_cache_dir = os.path.join(user_cache_dir("genomepy"), str(linux))
    os.makedirs(my_cache_dir, exist_ok=True)
    cache = Cache(directory=my_cache_dir, size_limit=1000000)
    test = ["a", "b", "c"]

    @cache.memoize(expire=10, tag="expensive_function")
    def expensive_function(data):
        return pd.DataFrame(data)

    # Get full function name https://github.com/grantjenks/python-diskcache/blob/master/diskcache/core.py
    def full_name(func):
        """Return full name of `func` by adding the module and function name."""
        return func.__module__ + "." + func.__qualname__

    # cache key tuple
    cache_key = (
        full_name(expensive_function),
        test,
        None,
    )

    # check that results before/after caching are identical
    expected = expensive_function(test)
    cached_data = cache.get(cache_key)
    assert cached_data.equals(expected), "Cached data does not match expected data"
    rmtree(my_cache_dir, ignore_errors=True)


# exceptions.py


def test_exceptions():
    with pytest.raises(genomepy.exceptions.GenomeDownloadError):
        raise genomepy.exceptions.GenomeDownloadError


# config.__init__.py


def test_config():
    config = norns.config("genomepy", default="config/default.yaml")
    assert len(config.keys()) == 5


def test_manage_config(capsys):
    # make a new config
    genomepy.config.manage_config("generate")
    captured = capsys.readouterr().out.strip()
    assert captured.startswith("Created config file")

    # check where it is found
    fname = os.path.expanduser("~/Library/Application Support/genomepy/genomepy.yaml")
    if linux:
        fname = os.path.expanduser("~/.config/genomepy/genomepy.yaml")
    genomepy.config.manage_config("file")
    captured = capsys.readouterr().out.strip()
    assert captured == fname

    # mess with the config
    with open(fname, "w") as f:
        print("bgzip: na", file=f)

    # show the mess
    genomepy.config.manage_config("show")
    captured = capsys.readouterr().out.strip()
    assert captured.startswith("bgzip: na")

    # make a new config
    genomepy.config.manage_config("generate")
    captured = capsys.readouterr().out.strip()
    assert captured.startswith("Created config file")

    # check if the mess was fixed
    genomepy.config.manage_config("show")
    captured = capsys.readouterr().out.strip()
    assert captured.startswith("bgzip: false")



import os.path
import re
import zlib
from urllib.request import urlopen

from loguru import logger

from genomepy.plugins import Plugin


class BlacklistPlugin(Plugin):
    # stanford_url = "http://mitra.stanford.edu/kundaje/akundaje/release/blacklists/"
    # encode_url = "https://www.encodeproject.org/files/"
    github_url = "https://github.com/Boyle-Lab/Blacklist/raw/master/lists/"
    http_dict = {
        "ce10": github_url + "ce10-blacklist.v2.bed.gz",
        "ce11": github_url + "ce11-blacklist.v2.bed.gz",
        "dm3": github_url + "dm3-blacklist.v2.bed.gz",
        "dm6": github_url + "dm6-blacklist.v2.bed.gz",
        "hg19": github_url + "hg19-blacklist.v2.bed.gz",
        "GRCh37": github_url + "hg19-blacklist.v2.bed.gz",
        "hg38": github_url + "hg38-blacklist.v2.bed.gz",
        "GRCh38": github_url + "hg38-blacklist.v2.bed.gz",
        "mm9": github_url + "Blacklist_v1/mm9-blacklist.bed.gz",  # no v2
        "mm10": github_url + "mm10-blacklist.v2.bed.gz",
        "GRCm38": github_url + "mm10-blacklist.v2.bed.gz",
        # for testing purposes
        "this was a triumph": "I'm making a note here: 'Huge success'",
    }

    def after_genome_download(self, genome, threads=1, force=False):
        fname = self.get_properties(genome)["blacklist"]
        if os.path.exists(fname) and not force:
            return

        link = self.http_dict.get(genome.name.split(".")[0])
        if link is None:
            logger.warning(f"No blacklist found for {genome.name}")
            return

        logger.info(f"Downloading blacklist {link}")
        try:
            response = urlopen(link)
            with open(fname, "wb") as bed:
                # unzip the response with some zlib magic
                unzipped = zlib.decompress(response.read(), 16 + zlib.MAX_WBITS)
                bed.write(unzipped)
        except Exception as e:
            logger.error(str(e))
            logger.error(f"Could not download blacklist file from {link}")

        # convert UCSC format to Ensembl/NCBI format
        if "1" in genome.sizes:
            os.rename(fname, fname + ".tmp")
            with open(fname + ".tmp") as old_bed, open(fname, "w") as new_bed:
                for line in old_bed:
                    line = line[3:]  # remove "chr" from the start of each line
                    new_bed.write(line)
            os.unlink(fname + ".tmp")

    def get_properties(self, genome):
        props = {"blacklist": re.sub(".fa(.gz)?$", ".blacklist.bed", genome.filename)}
        return props



"""
Global fixtures and functions for pytest
pytest can only share fixtures between modules if they are declared here.
"""
import logging
import os

import pytest
from loguru import logger

import genomepy.providers
from genomepy.providers.base import BaseProvider
from genomepy.providers.ensembl import EnsemblProvider
from genomepy.providers.gencode import GencodeProvider
from genomepy.providers.local import LocalProvider
from genomepy.providers.ncbi import NcbiProvider
from genomepy.providers.ucsc import UcscProvider
from genomepy.providers.url import UrlProvider


@pytest.fixture(scope="function")
def caplog(caplog):
    """Fixture is necessary to be able to check loguru log messages"""

    class PropogateHandler(logging.Handler):
        def emit(self, record):
            logging.getLogger(record.name).handle(record)

    handler_id = logger.add(PropogateHandler(), format="{message} {extra}")
    yield caplog
    logger.remove(handler_id)


def teardown(gprefix, skip=None):
    for ext in [
        "fa.fai",
        "fa.sizes",
        "gaps.bed",
        "fa.gz.fai",
        "fa.gz.sizes",
        "annotation.gtf",
        "annotation.bed",
    ]:
        if skip and ext in skip:
            continue
        file = gprefix + ext
        if os.path.exists(file):
            os.remove(file)
    gdir = os.path.dirname(gprefix)
    readme = os.path.join(gdir, "README.txt")
    if os.path.exists(readme):
        os.remove(readme)


@pytest.fixture(scope="function")
def small_genome():
    yield genomepy.Genome("tests/data/small_genome.fa.gz")

    teardown("tests/data/small_genome.")


@pytest.fixture(scope="function")
def gap_genome():
    yield genomepy.Genome("tests/data/gap.fa")

    teardown("tests/data/gap.")


@pytest.fixture(scope="function")
def annot():
    genome_file = "tests/data/regexp/regexp.fa"
    gtf_file = "tests/data/regexp/regexp.annotation.gtf"
    bed_file = "tests/data/regexp/regexp.annotation.bed"

    genomepy.Genome(genome_file)
    with open(gtf_file, "w") as f:
        f.write("# skip this line\n")
        f.write(
            """chrM\tvanHeeringen-lab\tNP_059343.1\t15307\t16448\t42\t+\t.\tattributes"""
        )
    with open(bed_file, "w") as f:
        f.write(
            """chrM\t15307\t16448\tNP_059343.1\t42\t+\t15307\t16448\t0\t1\t1141,\t0,"""
        )
    yield genomepy.Annotation("regexp", genomes_dir="tests/data")

    teardown("tests/data/regexp/regexp.")


def validate_annot(fname, ftype):
    """fname = path, ftype = 'bed' or 'gtf'."""
    assert os.path.exists(fname)
    columns = 12 if ftype == "bed" else 9
    start, end = (3, 4) if ftype == "gtf" else (1, 2)
    with open(fname, "r") as f:
        for line in f:
            if line.startswith("#"):
                continue
            vals = line.split("\t")
            assert columns == len(vals)
            int(vals[start]), int(vals[end])
            break


@pytest.fixture(scope="function")
def base():
    return BaseProvider()


@pytest.fixture(scope="function")
def ensembl():
    return EnsemblProvider()


@pytest.fixture(scope="function")
def ucsc():
    return UcscProvider()


@pytest.fixture(scope="function")
def gencode():
    return GencodeProvider()


@pytest.fixture(scope="function")
def ncbi():
    return NcbiProvider()


@pytest.fixture(scope="function")
def local():
    return LocalProvider()


@pytest.fixture(scope="function")
def url():
    return UrlProvider()


@pytest.fixture(scope="function")
def provider():
    return genomepy.Provider()



import pytest

import genomepy.plugins


@pytest.fixture
def deactivate_plugins():
    # save originally active plugins
    original_plugins = [p.name for p in genomepy.plugins.get_active_plugins()]
    # deactivate all plugins
    [genomepy.plugins.deactivate(p) for p in original_plugins]

    yield

    # reactivate original plugins
    [genomepy.plugins.activate(p) for p in original_plugins]


def test_plugin():
    p = genomepy.plugins.Plugin()
    assert p.name == "plugin"

    for class_method in [
        "activate",
        "deactivate",
        "after_genome_download",
        "get_properties",
    ]:
        assert class_method in dir(p)

    assert p.active is False
    p.activate()
    assert p.active is True


def test_convert(name="TestName", expected="test_name"):
    result = genomepy.plugins.convert(name)
    assert result == expected


def test_list_plugins():
    plugins = genomepy.plugins.list_plugins()
    expected = ["blacklist", "bowtie2", "bwa", "gmap", "hisat2", "minimap2", "star"]
    assert sorted(plugins) == sorted(expected)


def test_init_plugins():
    p = genomepy.plugins.init_plugins()
    assert isinstance(p, dict)

    plugins = list(p.keys())
    expected = ["blacklist", "bowtie2", "bwa", "gmap", "hisat2", "minimap2", "star"]
    assert sorted(plugins) == sorted(expected)


def test_get_active_plugins(deactivate_plugins):
    _ = deactivate_plugins

    # returns list of all active plugins
    genomepy.plugins.activate("bwa")
    active_plugins = genomepy.plugins.get_active_plugins()
    assert isinstance(active_plugins, list)
    assert ["bwa"] == [p.name for p in active_plugins]


def test_activate(deactivate_plugins):
    _ = deactivate_plugins

    assert len(genomepy.plugins.get_active_plugins()) == 0
    genomepy.plugins.activate("bwa")
    # don't crash activating active plugins
    genomepy.plugins.activate("bwa")
    assert len(genomepy.plugins.get_active_plugins()) == 1
    genomepy.plugins.deactivate("bwa")


def test_deactivate(deactivate_plugins):
    _ = deactivate_plugins

    genomepy.plugins.activate("bwa")
    assert len(genomepy.plugins.get_active_plugins()) == 1
    genomepy.plugins.deactivate("bwa")
    assert len(genomepy.plugins.get_active_plugins()) == 0
    # don't crash deactivating inactive plugins
    genomepy.plugins.deactivate("bwa")


def test_manage_plugins(capsys):
    genomepy.plugins.manage_plugins("enable", ["blacklist"])
    genomepy.plugins.manage_plugins("list")
    captured = capsys.readouterr().out.strip().split("\n")
    assert captured[2].startswith("blacklist")
    assert captured[2].endswith("*")
    # don't crash activating active plugins
    genomepy.plugins.manage_plugins("enable", ["blacklist"])

    genomepy.plugins.manage_plugins("disable", ["blacklist"])
    genomepy.plugins.manage_plugins("list")
    captured = capsys.readouterr().out.strip().split("\n")
    assert captured[3].startswith("blacklist"), captured
    assert not captured[3].endswith("*")
    # don't crash deactivating inactive plugins
    genomepy.plugins.manage_plugins("disable", ["blacklist"])

    with pytest.raises(ValueError):
        genomepy.plugins.manage_plugins("blurp")



"""as_seqdict function & overloads"""
import os
import re
import sys
from functools import singledispatch
from io import TextIOWrapper
from tempfile import NamedTemporaryFile

try:
    import pybedtools
except ImportError:
    pass

import numpy as np
import pyfaidx
from Bio.SeqIO.FastaIO import SimpleFastaParser  # noqa: this is biopython

from genomepy.genome import Genome

# Regular expression to check for region (chr:start-end or genome@chr:start-end)
region_p = re.compile(r"^[^@]+@([^\s]+):(\d+)-(\d+)$")


@singledispatch
def as_seqdict(
    to_convert, genome=None, minsize=None  # noqa: arguments used dispaced functions
):
    """
    Convert input to a dictionary with name as key and sequence as value.

    If the input contains genomic coordinates, the genome needs to be
    specified. If minsize is specified all sequences will be checked if they
    are not shorter than minsize. If regions (or a region file) are used as
    the input, the genome can optionally be specified in the region using the
    following format: genome@chrom:start-end.

    Current supported input types include:
    * FASTA, BED and region files.
    * List or numpy.ndarray of regions.
    * pyfaidx.Fasta object.
    * pybedtools.BedTool object.

    Parameters
    ----------
    to_convert : list, str, pyfaidx.Fasta or pybedtools.BedTool
        Input to convert to FASTA-like dictionary
    genome : str, optional
        Genomepy genome name.
    minsize : int or None, optional
        If specified, check if all sequences have at least size minsize.

    Returns
    -------
    dict
        sequence names as key and sequences as value.
    """
    raise NotImplementedError(f"Not implement for {type(to_convert)}")


def _check_minsize(fa, minsize):
    """
    Raise ValueError if there is any sequence that is shorter than minsize.
    If minsize is None the size will not be checked.
    """
    if minsize is None:
        return fa

    for name, seq in fa.items():
        if len(seq) < minsize:
            raise ValueError(f"sequence {name} is shorter than {minsize}")

    return fa


def _genomepy_convert(to_convert, genome, minsize=None):
    """
    Convert a variety of inputs using track2fasta().
    """
    if genome is None:
        raise ValueError("input file is not a FASTA file, need a genome!")

    g = Genome(genome)
    tmpfile = NamedTemporaryFile()
    g.track2fasta(to_convert, tmpfile.name)

    fa = as_seqdict(tmpfile.name)
    return _check_minsize(fa, minsize)


def _as_seqdict_genome_regions(regions, minsize=None):
    """
    Accepts list of regions where the genome is encoded in the region,
    using the genome@chrom:start-end format.
    """
    genomic_regions = {}
    for region in regions:
        genome, region = region.split("@")
        genomic_regions.setdefault(genome, []).append(region)

    # test if all genomes are installed
    for genome in genomic_regions:
        Genome(genome)

    tmpfa = NamedTemporaryFile(mode="w", delete=False)
    for genome, g_regions in genomic_regions.items():
        g = Genome(genome)

        fa = g.track2fasta(g_regions)

        for seq in fa:
            seq.name = f"{genome}@{seq.name}"
            print(seq.__repr__(), file=tmpfa)

    tmpfa.flush()

    # Open tempfile and restore original sequence order
    fa = as_seqdict(tmpfa.name)
    fa = {region: fa[region] for region in regions}
    return _check_minsize(fa, minsize)


@as_seqdict.register(list)
def _as_seqdict_list(to_convert, genome=None, minsize=None):
    """
    Accepts list of regions as input.
    """
    if region_p.match(to_convert[0]):
        return _as_seqdict_genome_regions(to_convert, minsize)

    return _genomepy_convert(to_convert, genome, minsize)


@as_seqdict.register(TextIOWrapper)
def _as_seqdict_file_object(
    to_convert, genome=None, minsize=None  # noqa: arguments used dispaced functions
):
    """
    Accepts file object as input, should be a FASTA file.
    """
    fa = {x: y for x, y in SimpleFastaParser(to_convert)}
    return _check_minsize(fa, minsize)


@as_seqdict.register(str)
def _as_seqdict_filename(to_convert, genome=None, minsize=None):
    """
    Accepts filename as input.
    """
    if not os.path.exists(to_convert):
        raise ValueError("Assuming filename, but it does not exist")

    f = open(to_convert)
    fa = as_seqdict(f)

    if any(fa):
        return _check_minsize(fa, minsize)

    with open(to_convert) as f:
        while True:
            line = f.readline()
            if line == "":
                break
            if not line.startswith("#"):
                break

        if line == "":
            raise IOError(f"empty file {to_convert}")

        if region_p.match(line.strip()):
            regions = [region.strip() for region in [line] + f.readlines()]
            return _as_seqdict_genome_regions(regions, minsize=None)

    # Biopython parser resulted in empty dict
    # Assuming it's a BED or region file
    return _genomepy_convert(to_convert, genome, minsize)


@as_seqdict.register(pyfaidx.Fasta)
def _as_seqdict_pyfaidx(
    to_convert, genome=None, minsize=None  # noqa: arguments used dispaced functions
):
    """
    Accepts pyfaidx.Fasta object as input.
    """
    fa = {k: str(v) for k, v in to_convert.items()}
    return _check_minsize(fa, minsize)


@as_seqdict.register(np.ndarray)
def _as_seqdict_array(to_convert, genome=None, minsize=None):
    """
    Accepts numpy.ndarray with regions as input.
    """
    return as_seqdict(list(to_convert), genome, minsize)


# pybedtools is loaded
if "pybedtools" in sys.modules:
    if hasattr(pybedtools, "BedTool"):

        @as_seqdict.register(pybedtools.bedtool.BedTool)
        def _as_seqdict_bedtool(to_convert, genome=None, minsize=None):
            """
            Accepts pybedtools.BedTool as input.
            """
            return _genomepy_convert(
                ["{}:{}-{}".format(*f[:3]) for f in to_convert], genome, minsize
            )



import pytest
import requests

import genomepy.providers.ensembl


def test_ensemblprovider(ensembl):
    assert ensembl.name == "Ensembl"
    assert ensembl.taxid_fields == ["taxonomy_id"]


def test_genome_info_tuple(ensembl):
    t = ensembl._genome_info_tuple("KH", size=True)
    assert isinstance(t, tuple)
    assert t[:-1] == (
        "KH",
        "GCA_000224145.1",
        7719,
        True,
        "Ciona intestinalis",
        115227500,
    )


def test_get_version(ensembl):
    v = ensembl.get_version(vertebrates=True)
    assert v.isnumeric()
    assert int(v) > 100

    v = ensembl.get_version()
    assert v.isnumeric()
    assert int(v) > 50

    v = ensembl.get_version(set_version=42)
    assert v.isnumeric()
    assert v == "42"


def test_get_division(ensembl):
    genome = ensembl.genomes["TAIR10"]
    division, is_vertebrate = genomepy.providers.ensembl.get_division(genome)
    assert division == "plants"
    assert is_vertebrate is False

    genome = ensembl.genomes["GRCh38.p13"]
    division, is_vertebrate = genomepy.providers.ensembl.get_division(genome)
    assert division == "vertebrates"
    assert is_vertebrate is True


def test_get_genome_download_link(ensembl):
    # non vertebrate: soft masked
    link = ensembl.get_genome_download_link("TAIR10", mask="soft", **{"version": 46})
    assert (
        link
        == "http://ftp.ensemblgenomes.org/pub/release-46/plants/"
        + "fasta/arabidopsis_thaliana/dna/Arabidopsis_thaliana.TAIR10.dna_sm.toplevel.fa.gz"
    )

    # non vertebrate: < release 30
    link = ensembl.get_genome_download_link("TAIR10", mask="soft", **{"version": 25})
    assert (
        link
        == "http://ftp.ensemblgenomes.org/pub/release-25/plants/"
        + "fasta/arabidopsis_thaliana/dna/Arabidopsis_thaliana.TAIR10.25.dna_sm.toplevel.fa.gz"
    )

    # non vertebrate: entry not using url_name (see issue #205)
    link = ensembl.get_genome_download_link("Aqu1", mask="soft", **{"version": 35})
    assert (
        link
        == "http://ftp.ensemblgenomes.org/pub/release-35/metazoa/"
        + "fasta/amphimedon_queenslandica/dna/Amphimedon_queenslandica.Aqu1.dna_sm.toplevel.fa.gz"
    )

    # vertebrate with primary assembly: unmasked
    link = ensembl.get_genome_download_link("GRCz11", mask="none", **{"version": 98})
    assert (
        link
        == "http://ftp.ensembl.org/pub/release-98/fasta/"
        + "danio_rerio/dna/Danio_rerio.GRCz11.dna.primary_assembly.fa.gz"
    )

    # vertebrate with primary assembly: hard masked and toplevel only
    link = ensembl.get_genome_download_link(
        "GRCz11", mask="hard", **{"version": 98, "toplevel": True}
    )
    assert (
        link
        == "http://ftp.ensembl.org/pub/release-98/fasta/"
        + "danio_rerio/dna/Danio_rerio.GRCz11.dna_rm.toplevel.fa.gz"
    )

    # vertebrate: latest version
    version = ensembl.get_version(True)
    link = ensembl.get_genome_download_link("GRCz11", **{"toplevel": True})
    expected_link = (
        f"http://ftp.ensembl.org/pub/release-{version}/"
        "fasta/danio_rerio/dna/Danio_rerio.GRCz11.dna_sm.toplevel.fa.gz"
    )
    assert link == expected_link


def test_get_annotation_download_links(ensembl):
    # non vertebrate
    links = ensembl.get_annotation_download_links("TAIR10", **{"version": 46})
    expected_link = (
        "http://ftp.ensemblgenomes.org/pub/release-46/plants/"
        "gtf/arabidopsis_thaliana/Arabidopsis_thaliana.TAIR10.46.gtf.gz"
    )
    assert links[0] == expected_link

    # vertebrate
    links = ensembl.get_annotation_download_links("GRCz11", **{"version": 98})
    expected_link = (
        "http://ftp.ensembl.org/pub/release-98/gtf/"
        "danio_rerio/Danio_rerio.GRCz11.98.gtf.gz"
    )
    assert links[0] == expected_link

    # non vertebrate: latest version
    version = ensembl.get_version(False)
    links = ensembl.get_annotation_download_links("TAIR10")
    expected_link = (
        f"http://ftp.ensemblgenomes.org/pub/release-{version}/plants/"
        f"gtf/arabidopsis_thaliana/Arabidopsis_thaliana.TAIR10.{version}.gtf.gz"
    )
    assert links[0] == expected_link

    # vertebrate: latest version
    version = ensembl.get_version(True)
    links = ensembl.get_annotation_download_links("GRCz11")
    expected_link = (
        f"http://ftp.ensembl.org/pub/release-{version}/"
        f"gtf/danio_rerio/Danio_rerio.GRCz11.{version}.gtf.gz"
    )
    assert links[0] == expected_link


def test_request_json():
    divisions = genomepy.providers.ensembl.request_json(
        "https://rest.ensembl.org/", "info/divisions?"
    )
    assert isinstance(divisions, list)
    assert "EnsemblVertebrates" in divisions

    # test not r.ok
    with pytest.raises(requests.exceptions.HTTPError):
        genomepy.providers.ensembl.request_json("https://rest.ensembl.org/", "error")


def test_get_genomes(ensembl):
    assert isinstance(ensembl.genomes, dict)
    assert "KH" in ensembl.genomes
    genome = ensembl.genomes["KH"]
    assert isinstance(genome, dict)
    for field in (
        ensembl.accession_fields + ensembl.taxid_fields + ensembl.description_fields
    ):
        assert field in genome
    assert genome["taxonomy_id"] == 7719



"""Provider class, modules & related functions"""
import os
from typing import Optional

import pandas as pd
from loguru import logger

from genomepy.files import read_readme
from genomepy.providers.ensembl import EnsemblProvider
from genomepy.providers.gencode import GencodeProvider
from genomepy.providers.local import LocalProvider
from genomepy.providers.ncbi import NcbiProvider, download_assembly_report
from genomepy.providers.ucsc import UcscProvider
from genomepy.providers.url import UrlProvider
from genomepy.utils import get_genomes_dir

__all__ = [
    "Provider",
    "list_online_providers",
    "list_providers",
    "search",
    "map_locations",
    "download_assembly_report",
    "online_providers",
]

PROVIDERS = {
    "gencode": GencodeProvider,
    "ensembl": EnsemblProvider,
    "ucsc": UcscProvider,
    "ncbi": NcbiProvider,
    "local": LocalProvider,
    "url": UrlProvider,
}


def create(name: str):
    """
    Create a provider based on the provider name.

    Adds additional method(s) to the instance.

    Parameters
    ----------
    name : str
        Name of the provider (e.g. UCSC, Ensembl, ...)

    Returns
    -------
    provider
        Provider instance
    """
    name = name.lower()
    if name not in PROVIDERS:
        options = "', '".join(list_providers())
        raise ValueError(f"Unknown provider '{name}'. Options: '{options}'")

    p = PROVIDERS[name]
    p.download_assembly_report = staticmethod(download_assembly_report)
    return p()


def list_providers():
    """
    List of providers genomepy supports

    Returns
    -------
    list
        names of providers
    """
    return [p.name for p in PROVIDERS.values()]


def list_online_providers():
    """
    List of providers genomepy supports *that are online right now*.

    Returns
    -------
    list
        names of online providers
    """
    return [p.name for p in PROVIDERS.values() if p.ping()]


def online_providers(provider: str = None):
    """
    Check if the provider can be reached, or any provider if none is specified.

    Parameters
    ----------
    provider : str, optional
        Only try to yield the specified provider.

    Yields
    ------
    provider
        Provider instances
    """
    providers = [provider] if provider else list_providers()
    for provider in providers:
        try:
            yield create(provider)
        except ConnectionError as e:
            logger.warning(str(e))


def search(term: str or int, provider: str = None, size=False):
    """
    Search for a genome.

    If provider is specified, search only that specific provider, else
    search all providers. Both the name and description are used for the
    search. Search term is case-insensitive and can contain regex.

    Parameters
    ----------
    term : str, int
        Search term, case-insensitive, allows regex.
    provider : str , optional
        Only search the specified provider (faster).
    size : bool, optional
        Show absolute genome size.

    Yields
    ------
    list
        genome name, provider and metadata
    """
    for p in online_providers(provider):
        for row in p.search(term, size):
            ret = list(row[:1]) + [p.name] + list(row[1:])
            yield ret


def map_locations(
    frm: str, to: str, genomes_dir: Optional[str] = None
) -> Optional[pd.DataFrame]:
    """
    Load chromosome mapping from one assembly to another using the
    NCBI assembly reports.

    Parameters
    ----------
    frm: str
        A local genomepy genome name
    to: str
        target provider (UCSC, Ensembl or NCBI)
    genomes_dir: str, optional
        The genomes directory to look for the genomes.
        Will search the default genomes_dir if left blank.

    Returns
    -------
    pandas.DataFrame
        Chromosome mapping.
    """
    to_provider = to.lower()
    if to_provider not in ["ucsc", "ncbi", "ensembl"]:
        raise ValueError(f"Genomepy can only map to NCBI, UCSC or Ensembl, not '{to}'.")

    genomes_dir = get_genomes_dir(genomes_dir)
    frm_readme = os.path.join(genomes_dir, frm, "README.txt")
    frm_asm_report = os.path.join(genomes_dir, frm, "assembly_report.txt")
    if not os.path.exists(frm_readme):
        raise FileNotFoundError(f"Cannot find {frm} in {genomes_dir}.")

    metadata, _ = read_readme(frm_readme)
    frm_provider = metadata.get("provider").lower()
    if frm_provider == to_provider:
        logger.warning(f"You are attempting to map {frm} from {to} to {to}.")
        return

    asm_acc = metadata.get("assembly_accession")
    if not os.path.exists(frm_asm_report):
        download_assembly_report(asm_acc, frm_asm_report)
    if not os.path.exists(frm_asm_report):
        logger.warning("Cannot map without an assembly report.")
        return

    asm_report = pd.read_csv(frm_asm_report, sep="\t", comment="#", dtype=str)
    asm_report["ensembl_name"] = asm_report["Sequence-Name"]
    asm_report["ncbi_name"] = asm_report["Sequence-Name"]
    asm_report["ucsc_name"] = asm_report["UCSC-style-name"]

    # for Ensembl, use GenBank names for the scaffolds
    asm_report.loc[
        asm_report["Sequence-Role"] != "assembled-molecule", "ensembl_name"
    ] = asm_report.loc[
        asm_report["Sequence-Role"] != "assembled-molecule", "GenBank-Accn"
    ]

    if "ucsc" in [frm_provider, to_provider] and list(
        asm_report["ucsc_name"].unique()
    ) == ["na"]:
        logger.warning("UCSC style names not available for this assembly.")
        return

    mapping = asm_report[[f"{frm_provider}_name", f"{to_provider}_name"]]
    mapping = mapping.dropna().drop_duplicates().set_index(f"{frm_provider}_name")
    return mapping


class Provider:
    """Find & instantiate providers"""

    list = staticmethod(list_providers)
    online = staticmethod(list_online_providers)
    create = staticmethod(create)



import os
import re
import subprocess as sp
from shutil import copyfile
from time import sleep

import pytest

import genomepy.utils
from genomepy.plugins import activate, deactivate, get_active_plugins, init_plugins
from genomepy.plugins.blacklist import BlacklistPlugin
from genomepy.plugins.bowtie2 import Bowtie2Plugin
from genomepy.plugins.bwa import BwaPlugin
from genomepy.plugins.gmap import GmapPlugin
from genomepy.plugins.hisat2 import Hisat2Plugin
from genomepy.plugins.minimap2 import Minimap2Plugin
from genomepy.plugins.star import StarPlugin
from tests import linux, travis


@pytest.fixture(autouse=True)
def activate_plugins():
    # save originally active plugins
    original_plugins = [p.name for p in get_active_plugins()]
    # activate all plugins
    [activate(p) for p in init_plugins()]

    yield

    # deactivate all plugins
    [deactivate(p) for p in init_plugins()]
    # reactivate original plugins
    [activate(p) for p in original_plugins]


@pytest.fixture(scope="function", params=["unzipped", "bgzipped"])
def genome(request):
    """Create a test genome and location"""
    name = "ce10"  # Use fake name for blacklist test
    fafile = "tests/data/small_genome.fa.gz"

    # setup test dir
    genomes_dir = os.path.join(os.getcwd(), ".genomepy_plugin_tests")
    genome_dir = os.path.join(genomes_dir, name)
    genomepy.utils.mkdir_p(genome_dir)
    fname = os.path.join(genome_dir, f"{name}.fa.gz")
    copyfile(fafile, fname)

    # unzip genome if required
    if request.param == "unzipped":
        sp.check_call(["gunzip", fname])

        # add annotation (for STAR and hisat2) for 1 of 2 tests
        gtf_file = "tests/data/ce10.annotation.gtf.gz"
        aname = os.path.join(genome_dir, f"{name}.annotation.gtf.gz")
        copyfile(gtf_file, aname)

    yield genomepy.Genome(name, genomes_dir=genomes_dir)

    # tear down test dir
    genomepy.utils.rm_rf(genomes_dir)


def dont_overwrite(p, genome, fname):
    t0 = os.path.getmtime(fname)
    # OSX rounds down getmtime to the second
    if not linux:
        sleep(1)
    p.after_genome_download(genome, force=False)
    t1 = os.path.getmtime(fname)
    assert t0 == t1


def test_blacklist(caplog, genome):
    """Create blacklist."""
    # independent of bgzipping
    # no need to check for both .fa and .fa.gz.
    if genome.filename.endswith(".fa.gz"):
        pass

    p = BlacklistPlugin()
    fname = re.sub(".fa(.gz)?$", ".blacklist.bed", genome.filename)

    # no blacklist found
    genome.name = "ce01"
    p.after_genome_download(genome, force=True)
    assert f"No blacklist found for {genome.name}" in caplog.text

    # error downloading blacklist
    genome.name = "this was a triumph"
    p.after_genome_download(genome, force=True)
    link = "I'm making a note here: 'Huge success'"
    assert f"Could not download blacklist file from {link}" in caplog.text

    # download blacklist
    genome.name = "ce10"
    p.after_genome_download(genome, force=True)
    assert "ce10-blacklist" in caplog.text
    assert os.path.exists(fname)
    with open(fname) as blacklist:
        for line in blacklist:
            assert line.startswith("chr")
            break
    # os.unlink(fname)

    # # download Ensembl/NCBI blacklist
    # genome.name = "GRCh38"
    # p.after_genome_download(genome, force=True)
    # assert "ENCFF356LFX/@@download/ENCFF356LFX.bed.gz" in caplog.text
    # with open(fname) as blacklist:
    #     for line in blacklist:
    #         assert not line.startswith("chr")
    #         break

    # don't overwrite
    dont_overwrite(p, genome, fname)
    os.unlink(fname)

    # genome.name = "ce10"


def test_bowtie2(genome, threads=2):
    """Create bowtie2 index."""
    # can work with bgzipped genomes natively,
    # no need to check for both .fa and .fa.gz.
    if genome.filename.endswith(".fa"):
        pass

    p = Bowtie2Plugin()
    p.after_genome_download(genome, threads=threads, force=True)

    dirname = os.path.dirname(genome.filename)
    index_dir = os.path.join(dirname, "index", "bowtie2")
    fname = os.path.join(index_dir, f"{genome.name}.1.bt2")
    assert os.path.exists(index_dir)
    assert os.path.exists(fname)


def test_bwa(genome, threads=2):
    """Create bwa index."""
    # can work with bgzipped genomes natively,
    # no need to check for both .fa and .fa.gz.
    if genome.filename.endswith(".fa"):
        pass

    p = BwaPlugin()
    p.after_genome_download(genome, threads=threads, force=True)

    dirname = os.path.dirname(genome.filename)
    index_dir = os.path.join(dirname, "index", "bwa")
    fname = os.path.join(index_dir, f"{genome.name}.fa.sa")
    assert os.path.exists(index_dir)
    assert os.path.exists(fname)


@pytest.mark.skipif(not travis, reason="slow")
def test_gmap(genome, threads=2):
    """Create gmap index."""
    p = GmapPlugin()
    p.after_genome_download(genome, threads=threads, force=True)

    dirname = os.path.dirname(genome.filename)
    index_dir = os.path.join(dirname, "index", "gmap")
    fname = os.path.join(index_dir, f"{genome.name}.maps")
    assert os.path.exists(index_dir)
    assert os.path.exists(fname)


def test_hisat2(caplog, genome, threads=2):
    """Create hisat2 index."""
    p = Hisat2Plugin()
    p.after_genome_download(genome, threads=threads, force=True)

    dirname = os.path.dirname(genome.filename)
    index_dir = os.path.join(dirname, "index", "hisat2")
    fname = os.path.join(index_dir, f"{genome.name}.1.ht2")
    assert os.path.exists(index_dir)
    assert os.path.exists(fname)

    if genome.annotation_gtf_file:
        # check if splice-aware index is generated
        assert os.path.exists(os.path.join(genome.genome_dir, "splice_sites.txt"))
        assert os.path.exists(os.path.join(genome.genome_dir, "exon_sites.txt"))
        # check if annotation file is still the same
        assert os.path.exists(genome.annotation_gtf_file)
        assert genome.annotation_gtf_file.endswith(".gtf.gz")
    else:
        assert "Creating Hisat2 index without annotation file." in caplog.text


def test_minimap2(genome, threads=2):
    """Create minimap2 index."""
    # can work with bgzipped genomes natively,
    # no need to check for both .fa and .fa.gz.
    if genome.filename.endswith(".fa"):
        pass

    p = Minimap2Plugin()
    p.after_genome_download(genome, threads=threads, force=True)

    dirname = os.path.dirname(genome.filename)
    index_dir = os.path.join(dirname, "index", "minimap2")
    fname = os.path.join(index_dir, f"{genome.name}.mmi")
    assert os.path.exists(index_dir)
    assert os.path.exists(fname)


@pytest.mark.skipif(not travis, reason="slow")
def test_star(caplog, genome, threads=2):
    """Create star index."""
    p = StarPlugin()
    p.after_genome_download(genome, threads=threads, force=True)

    dirname = os.path.dirname(genome.filename)
    index_dir = os.path.join(dirname, "index", "star")
    fname = os.path.join(index_dir, "SA")
    assert os.path.exists(index_dir)
    assert os.path.exists(fname)

    if genome.annotation_gtf_file:
        # check if splice-aware index is generated
        assert "Creating star index..." in caplog.text
        # check if annotation file is still the same
        assert os.path.exists(genome.annotation_gtf_file)
        assert genome.annotation_gtf_file.endswith(".gtf.gz")
    else:
        assert "Creating STAR index without annotation file." in caplog.text



import os

from genomepy.plugins import Plugin
from genomepy.utils import cmd_ok, mkdir_p, rm_rf, run_index_cmd


class BwaPlugin(Plugin):
    def after_genome_download(self, genome, threads=1, force=False):
        if not cmd_ok("bwa"):
            return

        # Create index dir
        index_dir = genome.plugin["bwa"]["index_dir"]
        index_name = genome.plugin["bwa"]["index_name"]
        if force:
            # Start from scratch
            rm_rf(index_dir)
        mkdir_p(index_dir)

        if not any(fname.endswith(".bwt") for fname in os.listdir(index_dir)):
            # Create index
            if not os.path.exists(index_name):
                os.symlink(genome.filename, index_name)
            cmd = f"bwa index {index_name}"
            run_index_cmd("bwa", cmd)

    def get_properties(self, genome):
        props = {
            "index_dir": os.path.join(os.path.dirname(genome.filename), "index", "bwa"),
            "index_name": os.path.join(
                os.path.dirname(genome.filename),
                "index",
                "bwa",
                f"{genome.name}.fa",
            ),
        }
        return props



"""Utility functions"""
import itertools
import os
import re
import shutil
import subprocess as sp
import sys
import time
from typing import Any, Optional
from urllib.request import urlopen

from loguru import logger

from genomepy.config import config


def cleanpath(path):
    """Expand any path input to a literal path output"""
    return os.path.abspath(  # expand relative paths ('./' and '../')
        os.path.expanduser(  # expand '~'
            os.path.expandvars(path)  # expand '$VARIABLES'
        )
    )


def mkdir_p(path):
    """'mkdir -p' in Python"""
    path = cleanpath(path)
    os.makedirs(path, exist_ok=True)


def rm_rf(path):
    """'rm -rf' in Python"""
    path = cleanpath(path)
    if os.path.isfile(path):
        try:
            os.unlink(path)
        except OSError:  # in case of NTFS related issues
            pass
    elif os.path.isdir(path):
        shutil.rmtree(path, ignore_errors=True)


def get_genomes_dir(genomes_dir: str = None, check_exist: Optional[bool] = True) -> str:
    """import genomes_dir if none is given, and check validity"""
    if not genomes_dir:
        # backwards compatibility for "genome_dir" (this fixes issue #87)
        genomes_dir = config.get("genomes_dir", config.get("genome_dir", None))
    if not genomes_dir:
        raise FileNotFoundError("Please provide or configure a genomes_dir")

    genomes_dir = cleanpath(genomes_dir)
    if not os.path.exists(genomes_dir) and check_exist:
        raise FileNotFoundError(f"Genomes_dir {genomes_dir} does not exist!")

    return genomes_dir


def cmd_ok(cmd) -> bool:
    """Returns True if cmd can be run."""
    try:
        sp.check_call(cmd, stderr=sp.PIPE, stdout=sp.PIPE)
    except sp.CalledProcessError:
        # bwa gives return code of 1 with no argument
        pass
    except FileNotFoundError:
        logger.error(f"{cmd} not found, skipping")
        return False
    return True


def run_index_cmd(name, cmd):
    """Run command, show errors if the returncode is non-zero."""
    logger.info(f"Creating {name} index...")
    p = sp.Popen(cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)

    # show a spinner while the command is running
    spinner = itertools.cycle(["-", "\\", "|", "/"])
    while p.poll() is None:
        sys.stdout.write("\r" + next(spinner))
        time.sleep(0.15)
        sys.stdout.flush()
    sys.stdout.write("\b")  # clear the spinner

    stdout, stderr = p.communicate()
    if p.returncode != 0:
        logger.error(
            f"Indexing failed\n"
            f"stdout: {stdout.decode('utf8')}\n"
            f"stderr: {stderr.decode('utf8')}"
        )


def get_genomename(name):
    """return the name of the genome without path or extensions"""
    name = os.path.basename(name)  # remove path
    name = re.split(r"\.fa|\.fna|\.annot", name)[0]  # remove extensions
    return name


def get_remotename(name):
    """try to get the name from the url"""
    name = name.split("/")[-1]  # remove path
    name = name.replace(".gz", "")  # remove .gz
    name = os.path.splitext(name)[0]  # remove .fa/.fna/.fasta etc
    # remove unwanted substrings from the name (ex: _genomes or .est_)
    unwanted = [
        "genome",
        "genomic",
        "sequence",
        "dna",
        "cds",
        "pep",
        "transcript",
        "EST",
        "toplevel",
        "primary",
        "assembly",
    ]
    spacers = r"( ?-?_?\.?)"
    for substring in unwanted:
        name = re.sub(
            f"{spacers}{substring}(s?){spacers}", "", name, flags=re.IGNORECASE
        )
    return name


def safe(name: Any) -> str:
    """Replace spaces with underscores."""
    return str(name).strip().replace(" ", "_")


def lower(string: Any) -> str:
    """safe(str).lower() for case-insensitive text comparisons"""
    return safe(string).lower()


def get_localname(name: Any, localname=None) -> str:
    """
    Returns the safe version of the given localname.

    If no localname is provided, return the safe version of the name.
    If the name is a working URL, return the safe version of the filename.
    """
    if localname:
        return safe(localname)

    # Remote file
    if try_except_pass((IOError, ValueError), urlopen, name):
        name = get_remotename(name)

    # Local file
    elif os.path.exists(name):
        name = get_genomename(name)

    return safe(name)


def try_except_pass(errors, func, *args, **kwargs):
    """
    try to return FUNC with ARGS, pass on ERRORS

    parameters
    ----------
    errors
      a single error, or a tuple of errors.

    func
      a function that takes args and kwargs
    """
    try:
        return func(*args, **kwargs)
    except errors:
        pass


def check_ucsc_tools(tools: list = None):
    """
    Check if UCSC tools for gene annotation conversion are installed.
    These tools are installed by Bioconda, but not by Pip.
    """
    # check only one tool for simplicity
    if tools is None:
        tools = ["genePredToGtf"]

    for tool in tools:
        if shutil.which(tool) is None:
            raise OSError(
                "Installing gene annotations requires missing UCSC tools. "
                "See https://github.com/vanheeringen-lab/genomepy#pip "
                "for download links and details."
            )



import os
from shutil import copyfile
from tempfile import TemporaryDirectory

import pandas as pd

import genomepy


def test_ncbiprovider(ncbi):
    assert ncbi.name == "NCBI"
    assert ncbi.taxid_fields == ["species_taxid", "taxid"]


def test_genome_info_tuple(ncbi):
    t = ncbi._genome_info_tuple("ASM2732v1", size=True)
    assert isinstance(t, tuple)
    assert t[:-1] == (
        "ASM2732v1",
        "GCF_000027325.1",
        2097,
        True,
        "Mycoplasmoides genitalium G37",
        580076,
    )


def test_get_genome_download_link(ncbi):
    link = ncbi.get_genome_download_link("ASM2732v1")
    assert (
        link
        == "https://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/027/"
        + "325/GCF_000027325.1_ASM2732v1/GCF_000027325.1_ASM2732v1_genomic.fna.gz"
    )


def test__post_process_download(ncbi):
    name = "ASM2732v1"
    localname = "tmp"
    out_dir = os.getcwd()
    with TemporaryDirectory(dir=out_dir) as tmpdir:
        # copy fa file for unmasking
        g = os.path.join(tmpdir, localname + ".fa")
        copyfile("tests/data/gap.fa", g)

        ncbi._post_process_download(name=name, fname=g, out_dir=tmpdir, mask="hard")
        assert os.path.exists(g)
        with open(g) as f:
            ln = 0
            for line in f:
                if line.startswith(">"):
                    # test sequence names (should remain identical here)
                    names = line[1:].strip().split(" ")
                    assert names[0] == names[1]
                elif ln == 9:
                    # test masking
                    assert line.strip() == "ANNNNNNNA"
                ln += 1

        asm_report = os.path.join(tmpdir, "assembly_report.txt")
        assert os.path.exists(asm_report)


def test_get_annotation_download_link(ncbi):
    links = ncbi.get_annotation_download_links("ASM2732v1")
    assert links[0].endswith("GCF_000027325.1_ASM2732v1_genomic.gff.gz")


def test__get_genomes(ncbi):
    assert isinstance(ncbi.genomes, dict)
    assert "ASM2732v1" in ncbi.genomes
    genome = ncbi.genomes["ASM2732v1"]
    assert isinstance(genome, dict)
    for field in ncbi.accession_fields + ncbi.taxid_fields + ncbi.description_fields:
        assert field in genome
    assert genome["species_taxid"] == "2097"
    assert genome["taxid"] == "243273"


def test_download_assembly_report():
    assembly_report = "tests/data/sacCer3/assembly_report.txt"
    genomepy.providers.download_assembly_report("GCA_000146045", assembly_report)
    report = pd.read_csv(assembly_report, sep="\t", comment="#")

    assert isinstance(report, pd.DataFrame)
    assert list(report.columns) == genomepy.providers.ncbi.ASM_FORMAT



import os
import subprocess as sp

from loguru import logger

from genomepy.files import extracted_file
from genomepy.plugins import Plugin
from genomepy.utils import cmd_ok, mkdir_p, rm_rf, run_index_cmd


class Hisat2Plugin(Plugin):
    def after_genome_download(self, genome, threads=1, force=False):
        index_name = genome.plugin["hisat2"]["index_name"]
        if not cmd_ok("hisat2-build") or (
            os.path.exists(f"{index_name}.1.ht2") and not force
        ):
            return

        index_dir = genome.plugin["hisat2"]["index_dir"]
        rm_rf(index_dir)
        mkdir_p(index_dir)

        # gunzip genome if bgzipped and return up-to-date genome name
        with extracted_file(genome.filename) as fname:

            # index command
            cmd = f"hisat2-build -p {threads} {fname} {index_name}"

            # if an annotation is present, generate a splice-aware index
            gtf_file = genome.annotation_gtf_file
            print(gtf_file)
            if gtf_file:
                with extracted_file(gtf_file) as _gtf_file:
                    # generate splice and exon site files to enhance indexing
                    hisat_path = (
                        sp.Popen("which hisat2", stdout=sp.PIPE, shell=True)
                        .stdout.read()
                        .decode("utf8")
                        .strip()
                    )
                    splice_script = hisat_path + "_extract_splice_sites.py"
                    splice_file = os.path.join(genome.genome_dir, "splice_sites.txt")
                    sp.check_call(
                        f"python3 {splice_script} {_gtf_file} > {splice_file}",
                        shell=True,
                    )

                    exon_script = hisat_path + "_extract_exons.py"
                    exon_file = os.path.join(genome.genome_dir, "exon_sites.txt")
                    sp.check_call(
                        f"python3 {exon_script} {_gtf_file} > {exon_file}", shell=True
                    )

                    # update index command with annotation
                    cmd += f" --ss {splice_file} --exon {exon_file}"
                    # Create index
                    run_index_cmd("hisat2", cmd)
            else:
                logger.info("Creating Hisat2 index without annotation file.")

                # Create index
                run_index_cmd("hisat2", cmd)
            print(gtf_file)

    def get_properties(self, genome):
        props = {
            "index_dir": os.path.join(
                os.path.dirname(genome.filename), "index", "hisat2"
            ),
            "index_name": os.path.join(
                os.path.dirname(genome.filename), "index", "hisat2", genome.name
            ),
        }
        return props



from tempfile import mkdtemp

import pytest

import genomepy
import genomepy.utils
from tests import travis

skip = True
if not skip:

    @pytest.mark.skipif(travis, reason="slow")
    def test_ucsc_human():
        """Test UCSC.

        Download human genome from UCSC and retrieve a
        specific sequence.
        """
        tmp = mkdtemp()
        genomepy.install_genome("hg38", "UCSC", genomes_dir=tmp)
        g = genomepy.Genome("hg38", genomes_dir=tmp)
        seq = g["chr6"][166168664:166168679]
        assert str(seq) == "CCTCCTCGCTCTCTT"
        genomepy.utils.rm_rf(tmp)

    @pytest.mark.skipif(travis, reason="slow")
    def test_ensembl_human():
        """Test Ensembl.

        Download human genome from Ensembl and retrieve a
        specific sequence.
        """
        tmp = mkdtemp()
        genomepy.install_genome("GRCh38.p13", "Ensembl", genomes_dir=tmp)
        g = genomepy.Genome("GRCh38.p13", genomes_dir=tmp)
        seq = g["6"][166168664:166168679]
        assert str(seq) == "CCTCCTCGCTCTCTT"
        genomepy.utils.rm_rf(tmp)

    @pytest.mark.skipif(travis, reason="slow")
    def test_ncbi_human():
        """Test NCBI.

        Download human genome from NCBI and retrieve a
        specific sequence.
        """
        tmp = mkdtemp()
        genomepy.install_genome("GRCh38.p13", "NCBI", genomes_dir=tmp)
        g = genomepy.Genome("GRCh38.p13", genomes_dir=tmp)
        seq = g["6"][166168664:166168679]
        assert str(seq) == "CCTCCTCGCTCTCTT"
        genomepy.utils.rm_rf(tmp)



import os
import re

from genomepy.exceptions import GenomeDownloadError
from genomepy.files import get_file_info
from genomepy.providers.base import BaseProvider
from genomepy.utils import cleanpath, get_genomename


class LocalProvider(BaseProvider):
    """
    Local genome provider.

    Give a local genome the genomepy treatment.
    """

    name = "Local"
    _cli_install_options = {
        "path_to_annotation": {
            "long": "path-to-annotation",
            "help": "link to the annotation file, required if this is not in the same directory as the fasta file",
            "default": None,
        },
    }

    def __init__(self):
        self.genomes = {}

    @staticmethod
    def ping():
        return True

    def genome_taxid(self, name):
        return

    def assembly_accession(self, name):
        return

    def search(self, term, size=False):
        """return an empty generator,
        same as if no genomes were found at the other providers"""
        yield from ()

    def _genome_info_tuple(self, name, size=False):
        return tuple()

    def _check_name(self, name):
        """check if genome name can be found for provider"""
        return name

    def get_genome_download_link(self, path, mask=None, **kwargs):
        path = cleanpath(path)
        if not os.path.exists(path):
            raise FileNotFoundError(f"Local path to genome does not exist: {path}")
        return path

    def get_annotation_download_link(self, name: str, **kwargs) -> str:
        """
        Return a filepath to a matching annotation.

        Parameters
        ----------
        name : str
            genome name
        **kwargs: dict, optional:
            path_to_annotation : direct path to the gene annotation

        Returns
        -------
        str
            path

        Raises
        ------
        GenomeDownloadError
            if no functional path was found
        """
        path = kwargs.get("path_to_annotation")
        if path:
            path = cleanpath(path)
            if not os.path.exists(path):
                raise FileNotFoundError(
                    f"Local path to annotation does not exist: {path}"
                )
            ext = get_file_info(path)[0]
            if ext not in [".gtf", ".gff", ".gff3", ".bed"]:
                raise TypeError(
                    "Only (gzipped) gtf, gff and bed files are supported.\n"
                )
            return path

        paths = self.get_annotation_download_links(name)
        if paths:
            return paths[0]

        raise GenomeDownloadError(
            f"No gene annotations found for {get_genomename(name)}.\n"
        )

    def get_annotation_download_links(self, name, **kwargs):
        """Returns all files containing both name and an annotation extension"""
        name = cleanpath(name)
        genome_dir = os.path.dirname(name)
        search_list = os.listdir(genome_dir)
        search_name = get_genomename(name)

        hits = []
        for ext in ["gtf", "gff", "gff3"]:
            # .*? = non greedy filler. (\.gz)? = optional .gz
            expr = rf"{search_name}.*?\.{ext}(\.gz)?"  # noqa: W605
            for line in search_list:
                hit = re.search(expr, line, flags=re.IGNORECASE)
                if hit:
                    hit = os.path.join(genome_dir, hit[0])
                    hits.append(hit)

        return hits



"""Utility functions with files"""
import gzip
import os
import re
import shutil
import subprocess as sp
import tarfile
from contextlib import contextmanager
from glob import glob
from tempfile import TemporaryDirectory, mkdtemp
from typing import Optional, Tuple, Union
from zipfile import ZipFile

from tqdm.auto import tqdm

from genomepy.utils import cleanpath, rm_rf


def parse_file(fpath, skip: Union[tuple, str] = "#"):
    """basic file parsing"""
    fpath = cleanpath(fpath)
    with open(fpath) as lines:
        for line in lines:
            line = line.strip()
            if line.startswith(skip):
                continue
            yield line


def read_readme(readme: str) -> Tuple[dict, list]:
    """
    Parse a readme file.

    Parameters
    ----------
    readme: str
        filename

    Returns
    -------
    tuple
        metadata : dict
            genome metadata

        lines: list
            non-metadata text (such as regex info)
    """
    metadata = {
        "name": "na",
        "provider": "na",
        "original name": "na",
        "original filename": "na",
        "assembly_accession": "na",
        "tax_id": "na",
        "mask": "na",
        "genome url": "na",
        "annotation url": "na",
        "sanitized annotation": "na",
        "genomepy version": "na",
        "date": "na",
    }
    lines = []

    if not os.path.exists(readme):
        return metadata, lines

    # if the readme exists, overwrite all metadata fields found
    with open(readme) as f:
        for line in f.readlines():
            if ": " in line:
                vals = line.strip().split(": ")
                metadata[vals[0].strip()] = (": ".join(vals[1:])).strip()
            else:
                line = line.strip("\n").strip(" ")
                # blank lines are allowed, but only one in a row
                if not (
                    line == ""
                    and len(lines) > 0
                    and lines[len(lines) - 1].strip() == ""
                ):
                    lines.append(line)

    return metadata, lines


def write_readme(readme: str, metadata: dict, lines: list = None):
    """
    Create a new readme file with supplied information.

    Parameters
    ----------
    readme: str
        filename
    metadata: dict
        dict to write
    lines: list, optional
        additional lines of text to append
    """
    with open(readme, "w") as f:
        for k, v in metadata.items():
            print(f"{k}: {v}", file=f)
        if lines:
            for line in lines:
                print(line, file=f)


def update_readme(readme: str, updated_metadata: dict = None, extra_lines: list = None):
    """
    Update a readme file with supplied information.

    Parameters
    ----------
    readme: str
        filename
    updated_metadata: dict
        dict with updated data
    extra_lines: list, optional
        additional lines of text to append
    """
    metadata, lines = read_readme(readme)
    if updated_metadata:
        metadata = {**metadata, **updated_metadata}
    if extra_lines:
        lines = lines + extra_lines
    write_readme(readme, metadata, lines)


@contextmanager
def extracted_file(fname: str):
    """Context manager to work with (b)gzipped file."""
    new_fname = extract_gzip(fname)
    gzipped = True
    if new_fname is None:
        new_fname = fname
        gzipped = False

    try:
        yield new_fname
    finally:
        if gzipped:
            try:
                bgzip_and_name(new_fname)
            except Exception:
                gzip_and_name(new_fname)


def extract_archive(
    fname: str, outfile: Optional[str] = None, concat: bool = False
) -> Union[str, None]:
    """
    Extract files from an archive.

    Archive may be a gzipped or bgzipped file (.gz), a zipped file (.zip)
    or a tarball (.tar.gz). Optionally, if multiple files are present
    they may be concatenated into one file.

    Parameters
    ----------
    fname: str
        path to file

    outfile : str, optional
        Name of the output file.

    concat : bool, optional
        If zip or tar.gz contains multiple files, concatenate them.

    Returns
    -------
    tuple
        fname, str
            up to date filename
    """
    if fname.endswith((".tgz", ".tar.gz")):
        return extract_tarball(fname, outfile=outfile)
    elif fname.endswith(".gz"):
        return extract_gzip(fname, outfile=outfile)
    elif fname.endswith(
        ".zip",
    ):
        return extract_zip(fname, outfile=outfile, concat=concat)


def extract_tarball(fname, outfile=None, concat=True) -> Union[str, None]:
    """Convert tar of multiple FASTAs to one file."""
    fnames = []
    # Extract files to temporary directory
    tmp_dir = mkdtemp(dir=os.path.dirname(outfile))
    with tarfile.open(fname) as tar:
        tar.extractall(path=tmp_dir)
    for root, _, files in os.walk(tmp_dir):
        fnames += [os.path.join(root, fname) for fname in files]

    if len(fnames) > 1 and not concat:
        raise ValueError("tarball contains multiple files, but concat not specified!")

    # Concatenate (also works woth one file)
    with open(outfile, "w") as out:
        for infile in fnames:
            for line in open(infile):
                out.write(line)

    rm_rf(tmp_dir)

    return outfile


def extract_gzip(fname: str, outfile: Optional[str] = None) -> Union[str, None]:
    """
    Gunzips the file if gzipped.

    Also works on bgzipped files.

    Parameters
    ----------
    fname: str
        path to file

    outfile : str, optional
        Name of the output file.

    Returns
    -------
    tuple
        fname, str
            up to date filename, or None if failed
    """
    if not outfile:
        outfile = fname[:-3]

    if fname.endswith(".gz"):
        with gzip.open(fname, "rb") as f_in:
            with open(outfile, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
        os.unlink(fname)
        return outfile


def extract_zip(
    fname: str, outfile: Optional[str] = None, concat: bool = False
) -> Union[str, None]:
    """
    Unzips the file if zipped.

    Parameters
    ----------
    fname: str
        path to file

    outfile : str, optional
        Name of the output file.

    concat : bool, optional
        If zip or tar.gz contains multiple files, concatenate them.
    Returns
    -------
    tuple
        fname, str
            up to date filename
    """
    if not outfile:
        outfile = fname[:-4]

    with ZipFile(fname, "r") as fzip, TemporaryDirectory() as tmpdir:
        fzip.extractall(path=tmpdir)
        fnames = glob(f"{tmpdir}/*")

        if len(fnames) > 1 and not concat:
            raise ValueError(
                "Downloaded zip file contains multiple files, but concat is not specified!"
            )

        # Concatenate (also works with one file)
        with open(outfile, "w") as out:
            for infile in fnames:
                for line in open(infile):
                    out.write(line)

    os.unlink(fname)

    return outfile


def gzip_and_name(fname, gzip_file=True) -> str:
    """
    Gzip file if requested.

    Parameters
    ----------
    fname: str
        path to file
    gzip_file: bool, optional
        whether to gzip the file

    Returns
    -------
    str
        up to date filename
    """
    if gzip_file:
        with open(fname, "rb") as f_in:
            with gzip.open(fname + ".gz", "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
        os.unlink(fname)
        fname += ".gz"
    return fname


def bgzip_and_name(fname, bgzip_file=True) -> str:
    """
    Bgzip file if requested.

    Parameters
    ----------
    fname: str
        path to file
    bgzip_file: bool, optional
        whether to gzip the file

    Returns
    -------
    str
        up to date filename
    """
    if bgzip_file:
        ret = sp.check_call(["bgzip", fname])
        fname += ".gz"
        if ret != 0:
            raise Exception(f"Error bgzipping genome {fname}. Is pysam installed?")
    return fname


def _open(fname: str, mode: Optional[str] = "r"):
    """
    Return a function to open a (gzipped) file.

    Parameters
    ----------
    fname: str
        (gzipped) file path
    mode: str
        (r)ead or (w)rite.

    Returns
    -------
    generator
        opened file
    """
    if mode not in ["r", "w"]:
        raise ValueError("mode must be either 'r' or 'w'.")

    if fname.endswith(".gz"):
        return gzip.open(fname, mode + "t")
    return open(fname, mode)


def get_file_info(fname) -> Tuple[str, bool]:
    """
    Returns the lower case file type of a file, and if it is (g)zipped

    Parameters
    ----------
    fname: str
        filename

    Returns
    -------
    tuple
        ext: str
            fname file type
        is_compressed: bool
            whether fname was (g)zipped
    """
    fname = fname.lower()
    is_compressed = False
    if fname.endswith((".tgz", ".tar.gz")):
        is_compressed = True
        fname = re.sub(r"\.(tgz|tar\.gz)$", "", fname)
    elif fname.endswith(".gz"):
        is_compressed = True
        fname = fname[:-3]
    elif fname.endswith(".zip"):
        is_compressed = True
        fname = fname[:-4]
    split = os.path.splitext(fname)
    return split[1], is_compressed


def glob_ext_files(dirname, ext="fa") -> list:
    """
    Return (gzipped) file names in directory containing the given extension.

    Parameters
    ----------
    dirname: str
        Directory name.

    ext: str , optional
        Filename extension (default: fa).

    Returns
    -------
    list
        file names
    """
    fnames = glob(os.path.join(dirname, f"*.{ext}*"))
    return [f for f in fnames if f.endswith((ext, f"{ext}.gz"))]


def _apply_fasta_regex_func(infa, regex_func, outfa=None):
    """
    filter a Fasta using the regex function.

    infa: path to genome fasta

    regex_func: a function that takes a contig header and returns a bool

    outfa: path to output fasta. If None, infa is overwritten

    returns a list of excluded contigs
    """
    # move the original file to a tmp folder
    out_dir = os.path.dirname(infa)
    tmp_dir = mkdtemp(dir=out_dir)
    old_fname = os.path.join(tmp_dir, "original") if outfa is None else infa
    new_fname = os.path.join(tmp_dir, "filtered")
    shutil.move(infa, old_fname)

    # perform the filtering
    excluded_contigs = []
    keep_contig = True
    with open(old_fname) as old, open(new_fname, "w") as new:
        for line in tqdm(old, desc="Filtering Fasta", unit_scale=1, unit=" lines"):
            if line[0] == ">":
                keep_contig = regex_func(line)
                if keep_contig is False:
                    excluded_contigs.append(line[1:].split(" ")[0].strip())
            if keep_contig:
                new.write(line)

    # move the filtered file to the original folder
    shutil.move(new_fname, outfa if outfa else infa)
    rm_rf(tmp_dir)

    return excluded_contigs


def filter_fasta(
    infa: str,
    outfa: str = None,
    regex: str = ".*",
    invert_match: Optional[bool] = False,
) -> list:
    """
    Filter fasta file based on regex.

    Parameters
    ----------
    infa : str
        Filename of the input fasta file.
    outfa : str, optional
        Filename of the output fasta file. If None, infa is overwritten.
    regex : str, optional
        Regular expression used for selecting sequences.
        Matches everything if left blank.
    invert_match : bool, optional
        Select all sequence *not* matching regex if set.

    Returns
    -------
    list
        removed contigs
    """
    pattern = re.compile(regex)

    def keep(header):
        return bool(pattern.search(header)) is not invert_match

    return _apply_fasta_regex_func(infa, keep, outfa)



"""Module-level functions"""
import os
import re
import sys
from tempfile import mkdtemp
from typing import Optional

import pyfaidx
from appdirs import user_config_dir

from genomepy.annotation import Annotation
from genomepy.config import config
from genomepy.exceptions import GenomeDownloadError
from genomepy.files import (
    _apply_fasta_regex_func,
    bgzip_and_name,
    glob_ext_files,
    gzip_and_name,
    read_readme,
    update_readme,
)
from genomepy.genome import Genome
from genomepy.online import check_url
from genomepy.plugins import get_active_plugins
from genomepy.providers import download_assembly_report, online_providers
from genomepy.utils import (
    cleanpath,
    get_genomes_dir,
    get_localname,
    mkdir_p,
    rm_rf,
    safe,
    try_except_pass,
)


def head_annotations(name: str, provider=None, n: int = 2):
    """
    Quickly inspect the metadata of each available annotation for the specified genome.

    For UCSC, up to 4 gene annotation styles are available:
    "ncbiRefSeq", "refGene", "ensGene", "knownGene" (respectively).

    For NCBI, the chromosome names are not yet sanitized.

    Parameters
    ----------
    name: str
        genome name
    provider: str, optional
        only search the specified provider for the genome name
    n: int, optional
        number of lines to show
    """
    for p in online_providers(provider):
        if name in p.genomes:
            tmp_dir = mkdtemp()
            p.head_annotation(name, genomes_dir=tmp_dir, n=n)
            rm_rf(tmp_dir)


def list_available_genomes(provider=None, size=False) -> list:
    """
    List all available genomes.

    Parameters
    ----------
    provider : str, optional
        List genomes from specific provider. Genomes from all
        providers will be returned if not specified.
    size : bool, optional
        Show absolute genome size.

    Yields
    ------
    list
        tuples with genome name, provider and metadata
    """
    for p in online_providers(provider):
        for row in p.list_available_genomes(size):
            yield list(row[:1]) + [p.name] + list(row[1:])


def list_installed_genomes(genomes_dir: str = None) -> list:
    """
    List all locally available genomes.

    Parameters
    ----------
    genomes_dir : str, optional
        Directory with genomes installed by genomepy.

    Returns
    -------
    list
        genome names
    """
    genomes_dir = get_genomes_dir(genomes_dir, check_exist=False)
    if not os.path.exists(genomes_dir):
        return []

    installed_genomes = []
    for subdir in sorted(os.listdir(genomes_dir)):
        if _is_genome_dir(os.path.join(genomes_dir, subdir)):
            installed_genomes.append(subdir)
    return installed_genomes


def install_genome(
    name: str,
    provider: Optional[str] = None,
    genomes_dir: Optional[str] = None,
    localname: Optional[str] = None,
    mask: Optional[str] = "soft",
    keep_alt: Optional[bool] = False,
    regex: Optional[str] = None,
    invert_match: Optional[bool] = False,
    bgzip: Optional[bool] = None,  # None -> check config. False -> dont check.
    annotation: Optional[bool] = False,
    only_annotation: Optional[bool] = False,
    skip_matching: Optional[bool] = False,
    skip_filter: Optional[bool] = False,
    threads: Optional[int] = 1,
    force: Optional[bool] = False,
    **kwargs: Optional[dict],
) -> Genome:
    """
    Install a genome (& gene annotation).

    Parameters
    ----------
    name : str
        Genome name

    provider : str , optional
        Provider name. will try Gencode, Ensembl, UCSC and NCBI (in that order) if not specified.

    genomes_dir : str , optional
        Where to create the output folder.

    localname : str , optional
        Custom name for this genome.

    mask : str , optional
        Genome masking of repetitive sequences. Options: hard/soft/none, default is soft.

    keep_alt : bool , optional
        Some genomes contain alternative regions. These regions cause issues with
        sequence alignment, as they are inherently duplications of the consensus regions.
        Set to true to keep these alternative regions.

    regex : str , optional
        Regular expression to select specific chromosome / scaffold names.

    invert_match : bool , optional
        Set to True to select all chromosomes that *don't* match the regex.

    bgzip : bool , optional
        If set to True the genome FASTA file will be compressed using bgzip,
        and gene annotation will be compressed with gzip.

    threads : int , optional
        Build genome index using multithreading (if supported). Default: lowest of 8/all threads.

    force : bool , optional
        Set to True to overwrite existing files.

    annotation : bool , optional
        If set to True, download gene annotation in BED and GTF format.

    only_annotation : bool , optional
        If set to True, only download the gene annotation files.

    skip_matching : bool , optional
        If set to True, contigs in the annotation not matching
        those in the genome will not be corrected.

    skip_filter : bool , optional
        If set to True, the gene annotations will not be filtered to match the genome contigs.

    kwargs : dict , optional
        Provider specific options.

        toplevel : bool , optional
            Ensembl only: Always download the toplevel genome. Ignores potential primary assembly.

        version : int , optional
            Ensembl only: Specify release version. Default is latest.

        to_annotation : text , optional
            URL only: direct link to annotation file.
            Required if this is not the same directory as the fasta.

        path_to_annotation : text, optional
            Local only: path to local annotation file.
            Required if this is not the same directory as the fasta.

    Returns
    -------
    Genome
        Genome class with the installed genome
    """
    name = safe(name)
    localname = get_localname(name, localname)
    genomes_dir = get_genomes_dir(genomes_dir, check_exist=False)
    out_dir = os.path.join(genomes_dir, localname)
    genome_file = os.path.join(out_dir, f"{localname}.fa")
    provider = _provider_selection(name, localname, genomes_dir, provider)

    # check which files need to be downloaded
    genome_found = _is_genome_dir(out_dir)
    download_genome = (
        genome_found is False or force is True
    ) and only_annotation is False
    annotation_found = bool(glob_ext_files(out_dir, "annotation.gtf")) and bool(
        glob_ext_files(out_dir, "annotation.bed")
    )
    download_annotation = (annotation_found is False or force is True) and any(
        [
            annotation,
            only_annotation,
            skip_matching,
            skip_filter,
            kwargs.get("to_annotation"),
            kwargs.get("path_to_annotation"),
            kwargs.get("ucsc_annotation_type"),
        ]
    )

    genome = None
    genome_downloaded = False
    if download_genome:
        if force:
            _delete_extensions(out_dir, ["fa", "fai"])
        provider.download_genome(
            name,
            genomes_dir,
            mask=mask,
            localname=localname,
            **kwargs,
        )
        genome_found = True
        genome_downloaded = True

        # Filter genome
        _filter_genome(genome_file, regex, invert_match, keep_alt)

        # Generates a Fasta object and the genome index, gaps and sizes files
        genome = Genome(localname, genomes_dir=genomes_dir)

        # Download the NCBI assembly report
        asm_report = os.path.join(out_dir, "assembly_report.txt")
        asm_acc = genome.assembly_accession
        if not os.path.exists(asm_report) and asm_acc != "na":
            download_assembly_report(asm_acc, asm_report)

        # Export installed genome(s)
        generate_env(genomes_dir=genomes_dir)

    annotation_downloaded = False
    if download_annotation:
        if force:
            _delete_extensions(out_dir, ["annotation.gtf", "annotation.bed"])
        provider.download_annotation(name, genomes_dir, localname=localname, **kwargs)
        annotation_downloaded = bool(
            glob_ext_files(out_dir, "annotation.gtf")
        ) and bool(glob_ext_files(out_dir, "annotation.bed"))

    if annotation_downloaded:
        annotation = Annotation(localname, genomes_dir=genomes_dir)
        if genome_found and not (skip_matching and skip_filter):
            annotation.sanitize(not skip_matching, not skip_filter, True)

    # Run active plugins (also if the genome was downloaded earlier)
    if genome_found:
        genome = genome if genome else Genome(localname, genomes_dir=genomes_dir)
        for plugin in get_active_plugins():
            plugin.after_genome_download(genome, threads, force)

    # zip files downloaded now
    if bgzip is True or (bgzip is None and config.get("bgzip")):
        if annotation_downloaded:
            gzip_and_name(annotation.annotation_gtf_file)
            gzip_and_name(annotation.annotation_bed_file)
        if genome_downloaded:
            bgzipped_genome = bgzip_and_name(genome.filename)

            # update references to bgzipped files
            genome = Genome(bgzipped_genome, genomes_dir=genomes_dir)
            generate_env(genomes_dir=genomes_dir)

    return genome


def generate_env(fname: str = "exports.txt", genomes_dir: str = None):
    """
    Generate file with exports.

    By default the export file generated is .config/genomepy/exports.txt.
    An alternative file name or file path is accepted too.

    Parameters
    ----------
    fname: str, optional
        Absolute path or name of the output file.

    genomes_dir: str, optional
        Directory with installed genomes to export.
    """
    fname1 = os.path.expanduser(os.path.expandvars(fname))
    fname2 = os.path.join(user_config_dir("genomepy"), fname)
    fname = fname1 if os.path.isabs(fname1) else fname2
    mkdir_p(os.path.dirname(fname))
    with open(fname, "w") as fout:
        for env in _generate_exports(genomes_dir):
            fout.write(f"{env}\n")


def _generate_exports(genomes_dir: str = None):
    """Print export commands for setting environment variables."""
    env = []
    for name in list_installed_genomes(genomes_dir):
        try:
            g = Genome(name, genomes_dir, build_index=False)
            env_name = re.sub(r"[^\w]+", "_", name).upper()
            env.append(f"export {env_name}={g.filename}")
        except (
            pyfaidx.FastaIndexingError,
            pyfaidx.IndexNotFoundError,
            FileNotFoundError,
        ):
            pass
    return env


def _lazy_provider_selection(name, provider=None):
    """return the first PROVIDER which has genome NAME"""
    providers = []
    for p in online_providers(provider):
        providers.append(p.name)
        if name in p.genomes:
            return p
        if p.name == "URL" and try_except_pass(ValueError, check_url, name):
            return p
        if p.name == "Local" and os.path.exists(cleanpath(name)):
            return p

    if len(providers):
        raise GenomeDownloadError(f"{name} not found on {', '.join(providers)}.")
    sys.exit(0)


def _provider_selection(name, localname, genomes_dir, provider=None):
    """
    Return a provider object

    First tries to return a specified provider,
    Second tries to return the provider from the README
    Third tries to return the first provider which has the genome (Ensembl>UCSC>NCBI)
    """
    readme = os.path.join(genomes_dir, localname, "README.txt")
    if provider is None and os.path.exists(readme):
        m, _ = read_readme(readme)
        p = m["provider"].lower()
        if p in ["ensembl", "ucsc", "ncbi"]:
            provider = p

    return _lazy_provider_selection(name, provider)


def _filter_genome(
    genome_file: str,
    regex: Optional[str] = None,
    invert_match: Optional[bool] = False,
    keep_alt: Optional[bool] = False,
):
    """
    Combine regex filters & document filtered contigs

    keep_alt : bool , optional
        Set to true to keep these alternative regions.

    regex : str , optional
        Regular expression to select specific chromosome / scaffold names.

    invert_match : bool , optional
        Set to True to select all chromosomes that don't match the regex.
    """
    if keep_alt is True and regex is None:
        return

    regex_func = _get_fasta_regex_func(regex, invert_match, keep_alt)
    excluded_contigs = _apply_fasta_regex_func(genome_file, regex_func)

    # document
    regex_line = "regex: "
    if keep_alt is False:
        regex_line += "'alt' (inverted match)" + (" and " if regex else "")
    if regex:
        regex_line += f"'{regex}'" + (" (inverted match)" if invert_match else "")
    lines = ["", regex_line, ""] + (
        [
            "The following contigs were filtered out of the genome:",
            f"{', '.join(excluded_contigs)}",
        ]
        if excluded_contigs
        else ["No contigs were removed."]
    )

    readme = os.path.join(os.path.dirname(genome_file), "README.txt")
    update_readme(readme, extra_lines=lines)


def _get_fasta_regex_func(
    regex: Optional[str] = None,
    invert_match: Optional[bool] = False,
    keep_alt: Optional[bool] = False,
):
    """
    returns a regex function that accepts a contig header and returns a bool to keep the contig or not.
    """
    # define filter functions
    if keep_alt is False:
        alt_pattern = re.compile("(alt)", re.I)  # case insensitive

        def alt_keep(header):
            return not bool(alt_pattern.search(header))

        keep = alt_keep  # rename in case there is only 1 filter function

    if regex is not None:
        re_pattern = re.compile(regex)

        def re_keep(header):
            return bool(re_pattern.search(header)) is not invert_match

        keep = re_keep  # rename in case there is only 1 filter function

    # combine filter functions?
    if regex is not None and keep_alt is False:

        def keep(header):
            return re_keep(header) and alt_keep(header)

    return keep  # noqa: IDE is confused


def _delete_extensions(directory: str, exts: list):
    """remove (gzipped) files in a directory matching any given extension"""
    for ext in exts:
        [rm_rf(f) for f in glob_ext_files(directory, ext)]


def _is_genome_dir(dirname):
    """
    Check if a directory contains a fasta file of the same name

    Parameters
    ----------
    dirname : str
        Directory name

    Returns
    ------
    bool
    """
    genome_file = os.path.join(dirname, f"{os.path.basename(dirname)}.fa")
    return os.path.exists(genome_file) or os.path.exists(f"{genome_file}.gz")



"""Functions to extract or convert genomic sequences"""
import os
import re
from bisect import bisect
from random import random

from loguru import logger
from pyfaidx import Sequence

from genomepy.files import parse_file


def track2fasta(
    self, track, fastafile=None, stranded=False, extend_up=0, extend_down=0  # noqa
):
    """
    Return a list of fasta sequences as Sequence objects
    as directed from the track(s).

    Parameters
    ----------
    track: list/region file/bed file
        region(s) you wish to translate to fasta.
        Example input files can be found in genomepy/tests/data/regions.*

    fastafile: bool , optional
        return Sequences as list or save to file? (default: list)

    stranded: bool , optional
        return sequences for both strands? Required BED6 (or higher) as input (default: False)

    extend_up: int , optional
        extend the sequences up? (command is strand sensitive, default: 0)

    extend_down: int , optional
        extend the sequences down? (command is strand sensitive, default: 0)
    """
    track_type = get_track_type(track)
    if track_type == "interval":
        seqqer = regions_to_seqs(
            self, track, extend_up=extend_up, extend_down=extend_down
        )
    else:
        seqqer = bed_to_seqs(
            self, track, stranded=stranded, extend_up=extend_up, extend_down=extend_down
        )

    if fastafile:
        with open(fastafile, "w") as fout:
            for seq in seqqer:
                fout.write(f"{seq.__repr__()}\n")
    else:
        return [seq for seq in seqqer]


def get_track_type(track):
    # region_p examples: ["chr1:123-456", "M:1-2"]
    region_p = re.compile(r"^([^\s]+):(\d+)-(\d+)")

    # check if track is a list (like), and if the first entry is an interval
    if not isinstance(track, (str, bytes)) and isinstance(track, (list, tuple)):
        if isinstance(track[0], (str, bytes)) and region_p.search(track[0]):
            return "interval"

    # check up to two (non-comment) lines for intervals (at the start)
    give_up = False
    for line in parse_file(track):
        if region_p.search(line):
            return "interval"
        if give_up:
            break
        give_up = True

    # default to BED format
    return "bed"


def bad_coords(region, track):
    chrom, coords = region.split(":")
    start, end = [int(c) for c in coords.split("-")]
    if end <= start:
        msg = f"Skipping region that ends before it starts: '{region}'"
        if isinstance(track, str):
            msg = f"Skipping region from '{os.path.basename(track)}' that ends before it starts: '{region}'"
        logger.warning(msg)
        return True
    return False


def region_to_seq(self, region, extend_up=0, extend_down=0):
    chrom, coords = region.strip().split(":")
    start, end = [int(c) for c in coords.split("-")]
    start += 1
    start -= extend_up
    end += extend_down
    seq = self.get_seq(chrom, start, end)
    return seq.seq


def regions_to_seqs(self, track, extend_up=0, extend_down=0):
    # if track is a file, loop over its lines
    lines = track
    if isinstance(track, str):
        lines = parse_file(track)

    for line in lines:
        name = line.split()[0]
        try:
            if bad_coords(name, track):
                continue
            seq = region_to_seq(self, name, extend_up, extend_down)
        except (ValueError, IndexError):
            msg = f"Skipping region that cannot be parsed: '{name}'"
            if isinstance(track, str):
                msg = f"Skipping region from '{os.path.basename(track)}' that cannot be parsed: '{name}'"
            logger.warning(msg)
            continue
        yield Sequence(name, seq)


def bed_to_seq(self, vals, stranded=False, extend_up=0, extend_down=0):
    chrom, start, end = str(vals[0]), int(vals[1]), int(vals[2])
    name = f"{chrom}:{start}-{end}"

    # there might be more...
    starts = [start]
    ends = [end]

    # BED4: add name column to name
    if len(vals) >= 4:
        name = " ".join((name, vals[3]))

    # BED5: check strandedness
    rc = False
    if stranded and len(vals) >= 6:
        rc = vals[5] == "-"

    # BED12: get all blocks
    if len(vals) >= 12:
        starts = [int(x) for x in vals[11].split(",")[:-1]]
        sizes = [int(x) for x in vals[10].split(",")[:-1]]
        starts = [start + x for x in starts]
        ends = [start + size for start, size in zip(starts, sizes)]
    # convert to 1-based counting
    starts = [start + 1 for start in starts]

    # extend
    if extend_up:
        if rc:
            ends[-1] += extend_up
        else:
            starts[0] -= extend_up
    if extend_down:
        if rc:
            starts[0] -= extend_down
        else:
            ends[-1] += extend_down

    intervals = zip(starts, ends)
    seq = self.get_spliced_seq(chrom, intervals, rc).seq
    return Sequence(name, seq)


def bed_to_seqs(self, track, stranded=False, extend_up=0, extend_down=0):
    for line in parse_file(track, skip=("#", "track")):
        vals = [val.strip() for val in line.split("\t")]
        try:
            name = f"{vals[0]}:{vals[1]}-{vals[2]}"
            if bad_coords(name, track):
                continue
            yield bed_to_seq(self, vals, stranded, extend_up, extend_down)
        except (ValueError, IndexError):
            msg = f"Skipping region from '{os.path.basename(track)}' that cannot be parsed: '{line}'"
            logger.warning(msg)
            continue


def get_random_sequences(
    self, n=10, length=200, chroms=None, max_n=0.1, outtype="list"  # noqa
):
    """
    Return random genomic sequences.

    Parameters
    ----------
    n : int , optional
        Number of sequences to return.

    length : int , optional
        Length of sequences to return.

    chroms : list , optional
        Return sequences only from these chromosomes.

    max_n : float , optional
        Maximum fraction of Ns.

    outtype : string , optional
        return the output as list or string.
        Options: "list" or "string", default: "list".

    Returns
    -------
    list
        coordinates as lists or strings:
        List with [chrom, start, end] genomic coordinates.
        String with "chrom:start-end" genomic coordinates
        (can be used as input for track2fasta).
    """
    if not chroms:
        chroms = self.keys()

    # dict of chromosome sizes after subtracting the number of Ns
    sizes = dict(
        [(chrom, len(self[chrom]) - self.gaps.get(chrom, 0)) for chrom in chroms]
    )

    # list of (tuples with) chromosomes and their size
    # (if that size is long enough for random sequence selection)
    lengths = [
        (sizes[x], x)
        for x in chroms
        if sizes[x] / len(self[x]) > 0.1 and sizes[x] > 10 * length
    ]
    if len(lengths) == 0:
        raise Exception("No contigs of sufficient size were found.")

    # random list of chromosomes from lengths (can have duplicates)
    chroms = weighted_selection(lengths, n)

    coords = []
    retries = 100
    cutoff = length * max_n
    for chrom in chroms:
        for _ in range(retries):
            start = int(random() * (sizes[chrom] - length))
            end = start + length
            count_n = self[chrom][start:end].seq.upper().count("N")
            if count_n <= cutoff:
                break
        else:
            raise Exception(
                f"Random subset ran {retries} times, "
                f"but could not find a sequence with less than {cutoff} N's in {chrom}.\n"
                "You can specify contigs using the CHROMS argument."
            )

        # list output example ["chr1", 123, 456]
        coords.append([chrom, start, end])

    if outtype != "list":
        # bed output example: "chr1:123-456"
        for i, region in enumerate(coords):
            coords[i] = [f"{region[0]}:{region[1]}-{region[2]}"]

    return coords


def weighted_selection(list_of_tuples, n):
    """
    Selects n random elements from a list of (weight, item) tuples.
    Based on code snippet by Nick Johnson
    """
    cuml = []
    items = []
    total_weight = 0.0
    for weight, item in list_of_tuples:
        total_weight += weight
        cuml.append(total_weight)
        items.append(item)

    return [items[bisect(cuml, random() * total_weight)] for _ in range(n)]



"""Mygene.info functions and methods"""
from typing import Iterable, Union

import mygene
import pandas as pd
from loguru import logger
from tqdm import tqdm

from genomepy.annotation.utils import _parse_annot
from genomepy.caching import cache_exp_other, disk_cache, lock


def _map_genes(
    self,  # noqa
    field: str,
    product: str = "protein",
    annot: Union[str, pd.DataFrame] = "bed",
) -> pd.DataFrame:
    """
    Use mygene.info to map gene identifiers to any specified `field`.

    Returns the dataframe with remapped "name" column.
    Drops missing identifiers.

    Parameters
    ----------
    annot: str or pd.Dataframe
        Annotation dataframe to map (a pandas dataframe or "bed").
        Is mapped to a column named "name" (required).
    field : str, optional
        Identifier for gene annotation. Uses mygene.info to map ids. Valid fields
        are: ensembl.gene, entrezgene, symbol, name, refseq, entrezgene. Note that
        refseq will return the protein refseq_id by default, use `product="rna"` to
        return the RNA refseq_id. Currently, mapping to Ensembl transcript ids is
        not supported.
    product : str, optional
        Either "protein" or "rna". Only used when `field="refseq"`

    Returns
    -------
    pandas.DataFrame
        remapped gene annotation
    """
    if not isinstance(self.tax_id, int):
        raise AttributeError(
            "A taxonomy identifier is required. "
            "You can set 'Annotation.tax_id' manually"
        )
    to, product = _parse_mygene_input(field, product)
    df = _parse_annot(self, annot)
    if df is None:
        raise ValueError("Argument 'annot' must be 'bed' or a pandas dataframe.")

    cols = df.columns  # starting columns
    if "name" not in cols:
        raise ValueError("Column 'name' is required to map to.")

    # remove version numbers from gene IDs
    split_id = df["name"].str.split(r"\.", expand=True)[0]
    df = df.assign(split_id=split_id.values)
    genes = sorted(set(split_id))

    result = query_mygene(genes, self.tax_id, field)
    # result = _query_mygene(self, genes, field=to)
    result = _filter_query(result)
    if len(result) == 0:
        logger.warning("Could not map using mygene.info")
        return pd.DataFrame()
    df = df.join(result, on="split_id")

    # Only in case of RefSeq annotation the product needs to be specified.
    if to == "refseq":
        to = f"{to}.translation.{product}"

    # Get rid of extra columns from query
    df = df.assign(name=df[to].values)  # df["name"] = df[to]
    df = df[cols].dropna()
    return df


@lock
@disk_cache.memoize(expire=cache_exp_other, tag="query_mygene")
def query_mygene(
    query: Iterable[str],
    tax_id: Union[str, int],
    field: str = "genomic_pos",
) -> pd.DataFrame:
    """
    Use mygene.info to map gene identifiers to another type.

    Parameters
    ----------
    query: iterable
        a list or list-like of gene identifiers
    tax_id: str or int
        Target genome taxonomy id
    field : str, optional
        Target identifier to map the query genes to. Valid fields
        are: ensembl.gene, entrezgene, symbol, name, refseq, entrezgene. Note that
        refseq will return the protein refseq_id by default, use `refseq.translation.rna` to
        return the RNA refseq_id. Currently, mapping to Ensembl transcript ids is
        not supported.

    Returns
    -------
    pandas.DataFrame
        mapped gene annotation.
    """
    field, _ = _parse_mygene_input(field)

    logger.info("Querying mygene.info...")
    query = list(query)
    query_len = len(query)
    batch_size = 1000  # same as mygene.info internal batch size
    it = range(0, query_len, batch_size)
    if query_len > batch_size:
        it = tqdm(it, unit=f"{batch_size} queries")

    result = pd.DataFrame()
    mg = mygene.MyGeneInfo()
    for i in it:
        _result = mg.querymany(
            query[i : i + batch_size],  # noqa
            scopes="symbol,name,ensembl.gene,entrezgene,ensembl.transcript,"
            "ensembl,accession.protein,accession.rna,other_names,alias",
            field=field,
            species=tax_id,
            as_dataframe=True,
            verbose=False,
        )
        result = pd.concat((result, _result))

    if "notfound" in result and result.shape[1] == 1:
        logger.warning("No matching genes found")

    return result


def _parse_mygene_input(field, product=None):
    if product is not None:
        product = product.lower()
        if product not in ["rna", "protein"]:
            raise ValueError("Argument 'product' should be either 'rna' or 'protein'.")

    field = field.lower()
    # see mg.MyGeneInfo().get_fields()
    allowed_fields = [
        "ensembl.gene",
        "entrezgene",
        "symbol",
        "name",
        "refseq",
        "genomic_pos",
    ]
    if product is None:
        allowed_fields.append("refseq.translation.rna")
    if field not in allowed_fields:
        raise ValueError(
            f"Argument 'field' should be either in {', '.join(allowed_fields)}"
        )

    return field, product


def _filter_query(query: pd.DataFrame) -> pd.DataFrame:
    """per queried gene, keep the best matching, non-NaN, mapping"""
    if "notfound" in query:
        query = query[query.notfound.isnull()]  # drop unmatched genes
        query = query.drop(columns=["notfound"])

    query = query.dropna()

    if "query" in query:
        query = query.groupby("query").first()  # already sorted by mapping score
        query = query.drop(columns=["_id", "_score"])
    return query



import argparse
import sys

from loguru import logger

from genomepy import Genome, install_genome


def parse_genome(auto_install=False, genomes_dir=None):
    """
    Argparse action for command-line genome option.

    Parameters
    ----------
        auto_install : bool, optional
            Install a genome if it's not found locally.
        genomes_dir : str, optional
            Directory to look for and/or install genomes.
    """

    class GenomeAction(argparse.Action):
        def __call__(self, parser, args, name, option_string=None):
            try:
                genome = Genome(name, genomes_dir=genomes_dir)
            except FileNotFoundError:
                logger.warning(f"Genome {name} not found!")
                if auto_install:
                    logger.info("Trying to install it automatically using genomepy...")
                    install_genome(name, annotation=True, genomes_dir=genomes_dir)
                    genome = Genome(name, genomes_dir=genomes_dir)
                else:
                    logger.info("You can install it using `genomepy install`.")
                    sys.exit(1)
            setattr(args, self.dest, genome)

    return GenomeAction
